
def retrieve_usgs_site_information_nldi(sSiteID):
    try:
        import requests
    except ImportError as e:
        raise ImportError(
            "The package 'requests' is required for this function to run.") from e
    
    nldi_site_url = f"https://labs.waterdata.usgs.gov/api/nldi/linked-data/nwissite/USGS-{sSiteID}"
    response = requests.get(nldi_site_url)   
    pData = response.json()    
    aLocation  = pData['features'][0]['geometry']['coordinates']
    dLon = aLocation[0]
    dLat = aLocation[1]
    
    return dLon, dLat


import numpy as np


def cgpercentiles(aData_in,
                  aPercentiles_in,
                  missing_value_in=None):
    """
    Calculates the percentiles of a given dataset.

    Args:
      aData_in (array-like): The input data.
      aPercentiles_in (array-like): The percentiles to calculate.
      missing_value_in (float, optional): The value to treat as missing. Defaults to None.

    Returns:
      list: The percentiles of the input data.
    """

    # Assume the worst.
    aData_in0 = np.array(aData_in)
    aData_in = aData_in0.flatten()
    aData_in = np.reshape(aData_in,  (aData_in.size))
    result = -1
    if aData_in.size == 0:
        print('Input data is required.')
        return -1

    aPercentiles_in = np.array(aPercentiles_in)
    aPercentiles_in = aPercentiles_in.flatten()
    if aPercentiles_in.size == 0:
        aPercentiles_in = np.array([0.25, 0.50, 0.75])

    if missing_value_in is not None:
        iFlag_missing_value = 1
        missing_value = missing_value_in
        dummy_index = np.where(aData_in == missing_value)
        aData_in[dummy_index] = np.nan
        good_index = np.where(np.isfinite(aData_in))
        aData_copy = aData_in[good_index]
    else:
        iFlag_missing_value = 0
        good_index = np.where(np.isfinite(aData_in))
        aData_copy = aData_in[good_index]

    num = aData_copy.size
    num_per = aPercentiles_in.size
    min_count = int(num/num_per)

    min_value = np.min(aData_copy)
    nan_index1 = np.where(aData_copy == min_value)

    max_value = np.max(aData_copy)
    nan_index2 = np.where(aData_copy == max_value)

    # np.count_nonzero(aData_copy == min_value)
    nan_count1 = nan_index1[0].size
    # np.count_nonzero(aData_copy == max_value)
    nan_count2 = nan_index2[0].size

    scenario = 4
    # there is no zero problem
    if ((min_value >= 0.0) or (max_value <= 0.0)):
        if ((nan_count1 > min_count) and (nan_count2 > min_count)):
            # both sides have problems
            scenario = 1
        else:
            if ((nan_count1 > min_count) and nan_count2 <= min_count):
                # min side has problem
                scenario = 2
            else:
                if ((nan_count2 > min_count) and (nan_count1 <= min_count)):
                    # max side has problem
                    scenario = 3
                else:
                    # everything is OK
                    scenario = 4

    else:

        nan_index3 = np.where(aData_copy == 0.0)
        nan_count3 = nan_index3[0].size

        if (nan_count3 >= min_count):
            # there is a probelm now

            if ((nan_count1 >= min_count) and (nan_count2 >= min_count)):
                # we have 3 problematic zones
                scenario = 5
            else:
                if ((nan_count1 >= min_count) and (nan_count2 <= min_count)):
                    # min side has problem and zero
                    scenario = 6
                else:
                    if ((nan_count2 >= min_count) and (nan_count1 <= min_count)):
                        # max side has problem and zero
                        scenario = 7
                    else:
                        # only zero problem
                        scenario = 8

        else:
            # not a 0.0 problem
            if ((nan_count1 >= min_count) and (nan_count2 >= min_count)):
                # we have 2 problematic zones
                scenario = 1
            else:
                if ((nan_count1 >= min_count) and (nan_count2 <= min_count)):
                    # min side has problem
                    scenario = 2
                else:
                    if ((nan_count2 >= min_count) and (nan_count1 <= min_count)):
                        # max side has problem
                        scenario = 3
                    else:
                        # only zero problem
                        scenario = 4

    # now

    if scenario == 1:
        good_index = np.where((aData_copy > min_value) &
                              (aData_copy < max_value))
        good_count = good_index[0].size
        if good_count <= num_per:
            return 0
        else:
            sample_count1 = np.ceil(
                float(good_count) / (num_per-1)) + 1  # why plus one?
            bad_copy1 = np.full(sample_count1, fill_value=min_value)
            sample_count2 = np.floor(float(good_count) / (num_per-1)) - 1
            bad_copy2 = np.full(sample_count2, fill_value=max_value)
            data_copy2 = [bad_copy1, bad_copy2, aData_copy[good_index]]

    else:
        if scenario == 2:
            good_index = np.where(aData_copy > min_value)
            good_count = good_index[0].size
            if good_count <= num_per:
                return 0
            else:
                sample_count = int(np.ceil(float(good_count) / (num_per)))
                bad_copy = np.full(sample_count, fill_value=min_value)
                # data_copy2 = [bad_copy,aData_copy[good_index]]
                data_copy2 = np.concatenate((bad_copy, aData_copy[good_index]))

        else:
            if scenario == 3:
                good_index = np.where(aData_copy < max_value)
                good_count = good_index[0].size
                if good_count <= num_per:
                    return 0
                else:
                    sample_count = int(np.ceil(float(good_count) / (num_per)))
                    bad_copy = np.full(sample_count, fill_value=max_value)
                    dummy = aData_copy[good_index]
                    data_copy2 = np.concatenate((bad_copy, dummy))

            else:
                if scenario == 4:
                    data_copy2 = aData_copy
                else:
                    if scenario == 5:
                        good_index = np.where((aData_copy > min_value) & (
                            aData_copy != 0.0) & (aData_copy < max_value))
                        good_count = good_index[0].size
                        if good_count <= num_per:
                            return 0
                        else:
                            sample_count1 = int(
                                np.ceil(float(good_count) / (num_per-2)) + 1)
                            bad_copy1 = np.full(
                                sample_count1, fill_value=min_value)
                            sample_count2 = int(
                                np.floor(float(good_count) / (num_per-2)) - 1)
                            bad_copy2 = np.full(sample_count2, fill_value=0.0)

                            sample_count3 = np.floor(
                                float(good_count) / (num_per-2)) - 1
                            bad_copy2 = np.full(sample_count3, fill_value=0.0)

                            # data_copy2 = [bad_copy1,bad_copy2, aData_copy[good_index]]
                            data_copy2 = np.concatenate(
                                (bad_copy1, bad_copy2, aData_copy[good_index]))

                    else:
                        if scenario == 6:
                            good_index = np.where(
                                (aData_copy > min_value) & (aData_copy != 0.0))
                            good_count = good_index[0].size
                            if good_count <= num_per:
                                return 0
                            else:
                                sample_count1 = int(
                                    np.ceil(float(good_count) / (num_per-1)) + 1)
                                bad_copy1 = np.full(
                                    sample_count1, fill_value=min_value)
                                sample_count2 = int(
                                    np.floor(float(good_count) / (num_per-1)) - 1)
                                bad_copy2 = np.full(
                                    sample_count2, fill_value=0.0)
                                # data_copy2 = [bad_copy1,bad_copy2, aData_copy[good_index]]
                                data_copy2 = np.concatenate(
                                    (bad_copy1, bad_copy2, aData_copy[good_index]))

                        else:
                            if scenario == 7:
                                good_index = np.where(
                                    (aData_copy != 0.0) & (aData_copy < max_value))
                                good_count = good_index[0].size
                                if good_count <= num_per:
                                    return 0
                                else:
                                    sample_count1 = int(
                                        np.ceil(float(good_count) / (num_per-1)) + 1)
                                    bad_copy1 = np.full(
                                        sample_count1, fill_value=0.0)
                                    sample_count2 = int(
                                        np.floor(float(good_count) / (num_per-1)) - 1)
                                    bad_copy2 = np.full(
                                        sample_count2, fill_value=max_value)
                                    # data_copy2 = [bad_copy1,bad_copy2, aData_copy[good_index]]
                                    data_copy2 = np.concatenate(
                                        (bad_copy1, bad_copy2, aData_copy[good_index]))

                            else:  # 8
                                good_index = np.where(aData_copy != 0.0)
                                good_count = good_index[0].size
                                if good_count <= num_per:
                                    return 0
                                else:
                                    sample_count = int(
                                        np.ceil(float(good_count) / (num_per)) - 1)
                                    bad_copy = np.full(
                                        sample_count, fill_value=0.0)
                                    # data_copy2 = [bad_copy,aData_copy[good_index]]
                                    data_copy2 = np.concatenate(
                                        (bad_copy, aData_copy[good_index]))

                                pass

    num = (data_copy2).size
    result = list()
    # Sort the data and find percentiles.
    for p in aPercentiles_in:
        a = np.percentile(data_copy2, p)
        result.append(a)

    return result



import os
import numpy as np
from osgeo import gdal, osr, gdalconst

from pyearth.gis.gdal.read.raster.gdal_get_raster_extent import gdal_get_raster_extent
from pyearth.gis.gdal.read.raster.gdal_get_raster_spatial_ref import gdal_get_raster_spatial_ref_wkt
from pyearth.gis.gdal.gdal_to_numpy_datatype import gdal_to_numpy_datatype
gdal.UseExceptions()    
#set default spatial reference
pSpatialRef_default = osr.SpatialReference()
pSpatialRef_default.ImportFromEPSG(4326)  # WGS84
sProjection_default = pSpatialRef_default.ExportToWkt()
pSpatialRef_default = None

def resample_raster(sFilename_in, sFilename_out, dResolution_x, dResolution_y,
                     pProjection_target_in = None, iData_type = gdalconst.GDT_Int16, 
                     sResampleAlg = 'MODE', dMissing_value_source= 255, dMissing_value_target = -9999):

    #check if the input raster exists
    if not os.path.exists(sFilename_in):
        print('Error: the input raster does not exist')
        return
    
    pDriver_tiff = gdal.GetDriverByName('GTiff')
    
    #get the spatial from the input raster
    pProjection_source = gdal_get_raster_spatial_ref_wkt(sFilename_in)

    #check whether it is wgs84 or not using project string
    pSpatialRef_source = osr.SpatialReference()
    pSpatialRef_source.ImportFromWkt(pProjection_source)
    if pProjection_source == sProjection_default:
        iFlag_wgs84_source = 1
    else:
        iFlag_wgs84_source = 0

    if pProjection_target_in is None:
        pProjection_target = pProjection_source
    else:
        pProjection_target = pProjection_target_in  

    
    if pProjection_target == sProjection_default:
        iFlag_wgs84_target = 1
    else:   
        iFlag_wgs84_target = 0
    
    if os.path.exists(sFilename_out):
        #check file size and remove if it is empty
        if os.path.getsize(sFilename_out) == 0:
            print('remove empty file: ' + sFilename_out)
            os.remove(sFilename_out) 
    else:
        print('create file: ' + sFilename_out)

    pSpatialRef_target = osr.SpatialReference()
    pSpatialRef_target.ImportFromWkt(pProjection_target)
    #use the same approach gire to define the extent
    #get the extent of raster geotiff file
    
    pDataset_in = gdal.Open(sFilename_in, gdal.GA_ReadOnly)

    if iFlag_wgs84_source == 1:       #from wgs84
        dLon_min, dLon_max, dLat_min, dLat_max = gdal_get_raster_extent(sFilename_in)
        if iFlag_wgs84_target == 1: #to wgs84
            #if the target is wgs84, we can use the same approach
            nleft  = np.floor(  (dLon_min - (-180)) /(dResolution_x)  )
            nright = np.ceil(  (dLon_max - (-180)) /(dResolution_x)  )
            ntop  = np.floor(  (90 - dLat_max) /(dResolution_y)  )
            nbot = np.ceil(  (90 - dLat_min) /(dResolution_y)  )
            nrow = int(nbot-ntop)
            ncolumn = int(nright - nleft)

            dMin_x = -180 + nleft*dResolution_x
            dMax_y = 90 - ntop*dResolution_y
        else: #to something else
            #if the target is not wgs84, we need to reproject the extent
            pTransform = osr.CoordinateTransformation(pSpatialRef_source, pSpatialRef_target)
            dX_min, dY_min, _ = pTransform.TransformPoint(dLon_min, dLat_min)
            dX_max, dY_max, _ = pTransform.TransformPoint(dLon_max, dLat_max)
            nrow = int((dY_max - dY_min)/dResolution_y)
            ncolumn = int((dX_max - dX_min)/dResolution_x)
            dMin_x = dX_min
            dMax_y = dY_max
    else:
        #if it is not wgs84, it has a projection already
        #get the extent of raster geotiff file
        dX_min, dX_max, dY_min, dY_max = gdal_get_raster_extent(sFilename_in)
        #if the projection is not the same, we need to reproject the extent
        if iFlag_wgs84_target == 1: #to wgs84
            pTransform = osr.CoordinateTransformation(pSpatialRef_source, pSpatialRef_target)
            dLon_min, dLat_min, _ = pTransform.TransformPoint(dX_min, dY_min)
            dLon_max, dLat_max, _ = pTransform.TransformPoint(dX_max, dY_max)
            nleft  = np.floor(  (dLon_min - (-180)) /(dResolution_x)  )
            nright = np.ceil(  (dLon_max - (-180)) /(dResolution_x)  )
            ntop  = np.floor(  (90 - dLat_max) /(dResolution_y)  )
            nbot = np.ceil(  (90 - dLat_min) /(dResolution_y)  )
            nrow = int(nbot-ntop)
            ncolumn = int(nright - nleft)
            dMin_x = -180 + nleft*dResolution_x
            dMax_y = 90 - ntop*dResolution_y
        else: #to something else, but could be the same or different
            if pProjection_source == pProjection_target:
                nrow = int((dY_max - dY_min)/dResolution_y)  
                ncolumn = int((dX_max - dX_min)/dResolution_x)   
                dMin_x = dX_min
                dMax_y = dY_max
            else: #not the same
                pTransform = osr.CoordinateTransformation(pSpatialRef_source, pSpatialRef_target)
                dX_min, dY_min, _ = pTransform.TransformPoint(dX_min, dY_min)
                dX_max, dY_max, _ = pTransform.TransformPoint(dX_max, dY_max)
                nrow = int((dY_max - dY_min)/dResolution_y) 
                ncolumn = int((dX_max - dX_min)/dResolution_x)    
                dMin_x = dX_min
                dMax_y = dY_max            

    iNewWidth = ncolumn
    iNewHeigh = nrow

    newGeoTransform = (dMin_x, dResolution_x, 0, dMax_y, 0, -dResolution_y)  
    pDataset_clip = pDriver_tiff.Create(sFilename_out, iNewWidth, iNewHeigh, 1, iData_type) #this data type may be provided as an input argument
    pDataset_clip.SetGeoTransform( newGeoTransform )
    pDataset_clip.SetProjection( pProjection_target) 
    
    pWrapOption = gdal.WarpOptions( cropToCutline=False, #could be true if vector file is provided
                                width=iNewWidth,   
                                    height=iNewHeigh,      
                                        dstSRS=pSpatialRef_target , format = 'MEM',
                                        resampleAlg=sResampleAlg ) #this resample algorithm may be provided as an input argument

    pDataset_clip_warped = gdal.Warp('', pDataset_in, options=pWrapOption)#gdal.Warp(sFilename_out, pDataset_in, options=pWrapOption)

    #convert the warped dataset to an array
    aData_clip = pDataset_clip_warped.ReadAsArray()
    #change the gdal data type to numpy data type
    iData_type_numpy = gdal_to_numpy_datatype(iData_type)

    aData_clip = aData_clip.astype(iData_type_numpy)
    aData_clip[aData_clip == dMissing_value_source] = dMissing_value_target
    pDataset_clip.GetRasterBand(1).WriteArray(aData_clip)        
    pDataset_clip = None
    pSpatialRef_source = None
    pSpatialRef_target = None
    return


import os
from osgeo import gdal


def gdal_write_geotiff_file(sFilename_in,
                            aData_in,
                            dPixelWidth_in,
                            dPixelHeight_in,
                            dOriginX_in,
                            dOriginY_in,
                            dMissing_value_in,
                            pSpatial_reference_in,
                            datatype=gdal.GDT_Float32):
    """
    Write a Geotiff standard format raster file

    Args:
        sFilename_in (string): The filename
        aData_in (numpy.array): The data
        dPixelWidth_in (float): The resolution
        dOriginX_in (float): The location of origin x
        dOriginY_in (float): The location of origin y
        dMissing_value_in (float): The missinge value
        pSpatial_reference_in (osr): The spatial reference

    Returns:
        Tuple: pGeotransform_out, pProjection_out
    """
    if os.path.exists(sFilename_in):
        os.remove(sFilename_in)
        pass

    sDriverName = 'GTiff'
    pDriver = gdal.GetDriverByName(sDriverName)

    if pDriver is None:
        print("%s pDriver not available.\n" % sDriverName)
    else:
        print("%s pDriver IS available.\n" % sDriverName)

    nrow, ncolumn = aData_in.shape
    nband = 1

    pDataset = pDriver.Create(
        sFilename_in,
        ncolumn,
        nrow,
        nband,
        datatype)

    pDataset.SetGeoTransform([
        dOriginX_in,    # 0
        dPixelWidth_in,  # 1
        0,                      # 2
        dOriginY_in,    # 3
        0,                      # 4
        dPixelHeight_in])

    pProjection = pSpatial_reference_in.ExportToPrettyWkt()
    pDataset.SetProjection(pProjection)

    pBand = pDataset.GetRasterBand(1)
    pBand.WriteArray(aData_in)
    pBand.SetNoDataValue(dMissing_value_in)

    pGeotransform_out = pDataset.GetGeoTransform()
    pProjection_out = pDataset.GetProjection()

    pDataset.FlushCache()  # Write to disk.
    pDriver = None
    pDataset = None
    pBand = None
    return sFilename_in, pGeotransform_out, pProjection_out


def gdal_write_geotiff_file_multiple_band(sFilename_in,
                                          aData_in,
                                          dPixelWidth_in,
                                          dPixelHeight_in,
                                          dOriginX_in,
                                          dOriginY_in,
                                          dMissing_value_in,
                                          pSpatial_reference_in,
                                          datatype=gdal.GDT_Float32):
    """
    Write a multi-band geotiff raster file

    Args:
        sFilename_in (string): The filename
        aData_in (numpy.array): The data
        dPixelWidth_in (float): The resolution
        dOriginX_in (float): The location of origin x
        dOriginY_in (float): The location of origin y
        dMissing_value_in (float): The missinge value
        pSpatial_reference_in (osr): The spatial reference

    Returns:
        Tuple: pGeotransform_out, pProjection_out
    """
    if os.path.exists(sFilename_in):
        os.remove(sFilename_in)
        pass

    sDriverName = 'GTiff'
    pDriver = gdal.GetDriverByName(sDriverName)

    if pDriver is None:
        print("%s pDriver not available.\n" % sDriverName)
    else:
        print("%s pDriver IS available.\n" % sDriverName)

    nband, nrow, ncolumn = aData_in.shape

    # Creates a new raster data source
    pDataset = pDriver.Create(sFilename_in,
                              ncolumn,
                              nrow,
                              nband,
                              datatype)

    # Write metadata

    pDataset.SetGeoTransform([dOriginX_in,
                              dPixelWidth_in,
                              0.0,
                              dOriginY_in,
                              0.0,
                              dPixelHeight_in])

    pProjection = pSpatial_reference_in.ExportToPrettyWkt()
    pDataset.SetProjection(pProjection)

    # Write raster datasets
    for iBand in range(nband):
        pBand = pDataset.GetRasterBand(iBand + 1)
        pBand.WriteArray(aData_in[iBand, :, :])
        pBand.SetNoDataValue(dMissing_value_in)

    pGeotransform_out = pDataset.GetGeoTransform()
    pProjection_out = pDataset.GetProjection()

    pDataset.FlushCache()  # Write to disk.
    pDriver = None
    pDataset = None
    pBand = None

    return sFilename_in, pGeotransform_out, pProjection_out



import os
def read_configuration_file(sFilename_configuration_in):
    """
    Read a text based configuration file

    Args:
        sFilename_configuration_in (string): The filename of the configuration file

    Returns:
        dict: The content in the configuration file
    """

    if os.path.exists(sFilename_configuration_in):
        pass
    else:
        print('The xml file does not exist!')
        return
 
    aConfig_out = {}
    ifs = open(sFilename_configuration_in, 'r')
    for sLine in ifs:
        sDummy = sLine.split(',')
        if (len(sDummy) == 2):
            print(sDummy)
            sKey = (sDummy[0]).strip()
            sValue = (sDummy[1]).strip()
            aConfig_out[sKey] = sValue
        else:
            pass
    ifs.close()
    return aConfig_out
 
    
    
    
    


import os
import sys
import numpy as np
from datetime import datetime
from osgeo import osr, ogr
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon
import cartopy as cpl
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
from pyearth.visual.formatter import OOMFormatter
pProjection = cpl.crs.PlateCarree()  # for latlon data only

def map_netcdf_file(sFilename_netcdf_in,
                    sVariable_in,
                    sFolder_output_in,
                    iFlag_monthly_in=None,
                    iFlag_daily_in=None,
                    iFlag_unstructured_in=None,                    
                    iFlag_scientific_notation_colorbar_in=None,
                    iFont_size_in=None,
                    sColormap_in=None,
                    sTitle_in=None,
                    iDPI_in=None,
                    iSize_x_in=None,
                    iSize_y_in=None,
                    dMissing_value_in=None,
                    dConvert_factor_in=None,
                    dData_max_in=None,
                    dData_min_in=None,
                    dResolution_x_in=None,
                    dResolution_y_in=None,
                    sExtend_in=None,
                    sFont_in=None,
                    sUnit_in=None,
                    aLegend_in=None,
                    aExtent_in=None,
                    pProjection_map_in=None,
                    pBoundary_in=None):
    """map a netcdf file using a variable

    Args:
        sFilename_netcdf_in (_type_): _description_
        sVariable_in (_type_): _description_
        sFolder_output_in (_type_): _description_
        iFlag_unstructured_in (_type_, optional): _description_. Defaults to None.
        iFlag_scientific_notation_colorbar_in (_type_, optional): _description_. Defaults to None.
        iFont_size_in (_type_, optional): _description_. Defaults to None.
        sColormap_in (_type_, optional): _description_. Defaults to None.
        sTitle_in (_type_, optional): _description_. Defaults to None.
        iDPI_in (_type_, optional): _description_. Defaults to None.
        iSize_x_in (_type_, optional): _description_. Defaults to None.
        iSize_y_in (_type_, optional): _description_. Defaults to None.
        dMissing_value_in (_type_, optional): _description_. Defaults to None.
        dConvert_factor_in (_type_, optional): _description_. Defaults to None.
        dData_max_in (_type_, optional): _description_. Defaults to None.
        dData_min_in (_type_, optional): _description_. Defaults to None.
        dResolution_x_in (_type_, optional): _description_. Defaults to None.
        dResolution_y_in (_type_, optional): _description_. Defaults to None.
        sExtend_in (_type_, optional): _description_. Defaults to None.
        sFont_in (_type_, optional): _description_. Defaults to None.
        sUnit_in (_type_, optional): _description_. Defaults to None.
        aLegend_in (_type_, optional): _description_. Defaults to None.
        aExtent_in (_type_, optional): _description_. Defaults to None.
        pProjection_map_in (_type_, optional): _description_. Defaults to None.
        pBoundary_in (_type_, optional): _description_. Defaults to None.

    Raises:
        ImportError: _description_
    """

    try:
        import netCDF4 as nc
    except ImportError as e:
        raise ImportError(
            "The package 'netCDF4' is required for this function to run.") from e

    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 100

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 8

    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 8

    if iFont_size_in is not None:
        iFont_size = iFont_size_in
    else:
        iFont_size = 12

    if iFlag_monthly_in is not None:
        iFlag_monthly = iFlag_monthly_in
    else:
        iFlag_monthly = 0

    if iFlag_daily_in is not None:
        iFlag_daily = iFlag_daily_in
    else:
        iFlag_daily = 0

    if dMissing_value_in is not None:
        dMissing_value = dMissing_value_in
    else:
        dMissing_value = -9999

    if dData_min_in is not None:
        iFlag_data_min = 1
        dValue_min = dData_min_in
    else:
        iFlag_data_min = 0
        pass

    if dData_max_in is not None:
        iFlag_data_max = 1
        dValue_max = dData_max_in
    else:
        iFlag_data_max = 0
        pass

    if dConvert_factor_in is not None:
        dConvert_factor = dConvert_factor_in
    else:
        dConvert_factor = 1.0

    if iFlag_scientific_notation_colorbar_in is not None:
        iFlag_scientific_notation_colorbar = iFlag_scientific_notation_colorbar_in
    else:
        iFlag_scientific_notation_colorbar = 0

    if sColormap_in is not None:
        sColormap = sColormap_in
    else:
        sColormap = 'rainbow'

    if sTitle_in is not None:
        sTitle = sTitle_in
        iFlag_title = 1
    else:
        iFlag_title = 0
        sTitle = ''

    if sExtend_in is not None:
        sExtend = sExtend_in
    else:
        sExtend = 'max'

    if sUnit_in is not None:
        sUnit = sUnit_in
    else:
        sUnit = ''

    if sFont_in is not None:
        sFont = sFont_in
    else:
        sFont = "Times New Roman"

    plt.rcParams["font.family"] = sFont

    sVariable = sVariable_in

    if pBoundary_in is None:
        pBoundary = None
    else:
        # for the reason that a geometry object will be crash if the associated dataset is closed, we must pass wkt string
        # https://gdal.org/api/python_gotchas.html
        pBoundary = ogr.CreateGeometryFromWkt(pBoundary_in)
        # pBoundary = pBoundary_in #only works for shapely geometry object
        pBoundary_box = pBoundary.GetEnvelope()

    cmap = mpl.cm.get_cmap(sColormap)

    # how can we define the boundary of the map?
    # if the boundary is not defined, we will use the boundary of the data
    # if the boundary is defined, we will use the boundary of the data

    # there are two ways to define the max and min of the data
    # the fist way is to use the max and min of the data,
    # this approach also have two sub-approaches: whole data, or data within the boundary
    # the second way is user defined max and min of the data

    pSrs = osr.SpatialReference()
    pSrs.ImportFromEPSG(4326)    # WGS84 lat/lon

    try:
        pDatasets_in = nc.Dataset(sFilename_netcdf_in, 'r')
        pDatasets_in.set_auto_mask(False)

        #get the year information from the filename 
        sYear = sFilename_netcdf_in[-7:-3]
        iYear = int(sYear)
        # get all the dimensions
        
        for sKey, aValue in pDatasets_in.variables.items():    
            if sKey == 'time':
                aTime = aValue
                nTime = len(aTime)

            if sKey == 'lat':
                aLatitude = aValue[:]
                
            if sKey == 'lon':
                dummy = aValue[:]
                if np.min(dummy) < 0:
                    aLongitude = dummy
                    iFlag_cutoff = 0
                else:
                    aLongitude = dummy - 180
                    iFlag_cutoff = 1

            if sKey == sVariable:
                aData = aValue[:] * dConvert_factor
                
                pass

        
        
        # deal with the time
        print('Extracting data ...')

        if iFlag_unstructured_in is not None:
            # unstructured grid
            pass
        else:
            # regular grid
            nVertex = 4
            dLongitude_min = np.min(aLongitude)
            dLongitude_max = np.max(aLongitude)
            dLatitude_min = np.min(aLatitude)
            dLatitude_max = np.max(aLatitude)

            if pProjection_map_in is not None:
                pProjection_map = pProjection_map_in
            else:
                if pBoundary_in is None:
                    pProjection_map = cpl.crs.Orthographic(central_longitude=0.50*(dLongitude_min+dLongitude_max),
                                                        central_latitude=0.50*(dLatitude_min+dLatitude_max), globe=None)
                else:
                    # get the center of the boundary
                    centroid = pBoundary.Centroid()
                    # Extract coordinates
                    centroid_x = centroid.GetX()
                    centroid_y = centroid.GetY()
                    pProjection_map = cpl.crs.Orthographic(central_longitude=centroid_x,
                                                        central_latitude=centroid_y, globe=None)

                    pass

            ncolumn = len(aLongitude)
            nrow = len(aLatitude)
            if iFlag_cutoff == 1:
                #shift the data
                shift = int(ncolumn/2)
                aData = np.roll(aData, shift, axis=2)                   
                pass
            
            if dResolution_x_in is not None:
                dResolution_x = dResolution_x_in
            else:
                dResolution_x = (dLongitude_max - dLongitude_min) / (ncolumn - 1)
            
            if dResolution_y_in is not None:
                dResolution_y = dResolution_y_in
            else:
                dResolution_y = (dLatitude_max - dLatitude_min) / (nrow - 1)

            # get max and min of the data
            print(dResolution_x, dResolution_y)
            # Pre-calculate half resolutions
            half_res_x = dResolution_x / 2
            half_res_y = dResolution_y / 2


            #predefine a lat-lon meshgrid using the center of the cell
            aLongitude_center = np.arange(ncolumn) * dResolution_x + dLongitude_min + half_res_x
            aLatitude_center = np.arange(nrow) * dResolution_y + dLatitude_min + half_res_y

            aLongitude_grid, aLatitude_grid = np.meshgrid(aLongitude_center, aLatitude_center)
            # Create a mask for cells within the boundary
            aMask = np.ones((nrow, ncolumn), dtype=bool)
            if pBoundary_in is not None:
                for i in range(nrow):
                    for j in range(ncolumn):
                        pPoint = ogr.Geometry(ogr.wkbPoint)
                        pPoint.AddPoint(aLongitude_grid[i, j], aLatitude_grid[i, j])
                        if not pPoint.Within(pBoundary):
                            aMask[i, j] = False

            #find min and max of the data
            if iFlag_data_max == 1 and iFlag_data_min == 1:
                aData_dump = list()
                for i in range(0, nrow):
                    for j in range(0, ncolumn):
                        if not aMask[i, j]:
                            continue
                        aData_dump.append(aData[:, i, j])
                        pass
                aData_dump = np.array(aData_dump)
                print(np.max(aData_dump))
                pass
            else:
                if pBoundary_in is not None:
                    aData_dump = list()
                    for i in range(0, nrow):
                        for j in range(0, ncolumn):
                            if not aMask[i, j]:
                                continue

                            aData_dump.append(aData[:, i, j])
                            pass

                    aData_dump = np.array(aData_dump)
                    if iFlag_data_max == 0:
                        dValue_max = np.max(aData_dump)
                    if iFlag_data_min == 0:
                        dValue_min = np.min(aData_dump)
                else:
                    if iFlag_data_max == 0:
                        dValue_max = np.max(aData)

                    if iFlag_data_min == 0:
                        dValue_min = np.min(aData)

            print(dValue_min, dValue_max)
            #flush the print buffer 
            
            sys.stdout.flush()
            
            # the bounding box should be applied to cell center
            
            #use the first day as index 
            pDate0 = datetime(iYear, 1, 1)
            if iFlag_monthly == 1:
                for iMonth in range(1,13):
                    sMonth = '{:02d}'.format(iMonth)
                    fig = plt.figure(dpi=iDPI)
                    fig.set_figwidth(iSize_x)
                    fig.set_figheight(iSize_y)
                    ax = fig.add_axes([0.08, 0.1, 0.62, 0.7],
                                      projection=pProjection_map)
                    ax.set_global()
                    # redefine the filename and title
                    sFilename_out = os.path.join(
                        sFolder_output_in, sVariable + '_' + sMonth + '.png')

                    #get the start and end index 
                    pDate_start = datetime(iYear, iMonth, 1)
                    if iMonth == 12:
                        pDate_end = datetime(iYear+1, 1, 1)
                    else:
                        pDate_end = datetime(iYear, iMonth+1, 1)

                    iStart = (pDate_start - pDate0).days
                    iEnd = (pDate_end - pDate0).days
                    
                    aPolygon = []
                    aColor = []
                    aData_dummy = []
                    for i in range(0, nrow):
                        for j in range(0, ncolumn):
                            if not aMask[i, j]:
                                continue

                            dLongitude_cell_center = aLongitude_grid[i, j]
                            dLatitude_cell_center = aLatitude_grid[i, j]
                            data_dummy = aData[iStart:iEnd, i, j]
                            dValue = np.mean(data_dummy)
                            aData_dummy.append(dValue)

                            # get its cell shape
                            aCoords_gcs = np.full((nVertex, 2), np.nan)

                            # Define coordinates for each vertex
                            coords = [(dLongitude_cell_center - half_res_x, dLatitude_cell_center - half_res_y),
                                      (dLongitude_cell_center + half_res_x, dLatitude_cell_center - half_res_y),
                                      (dLongitude_cell_center + half_res_x, dLatitude_cell_center + half_res_y),
                                      (dLongitude_cell_center - half_res_x, dLatitude_cell_center + half_res_y)]

                            # Fill aCoords_gcs using a loop
                            for idx, (x, y) in enumerate(coords):
                                aCoords_gcs[idx, :] = x, y

                            #dealing with missing value
                            if dValue > dValue_min: 
                                iColor_index = np.clip(((dValue - dValue_min) / (dValue_max - dValue_min) * 255).astype(int), 0, 255)
                                # pick color from colormap       
                                aColor.append(cmap(iColor_index))
                                aPolygon.append(aCoords_gcs[:, 0:2])

                    print(np.max(aData_dummy))
                    # add extent
                    # Create a PatchCollection with the polygons
                    aPatch = [Polygon(poly, closed=True) for poly in aPolygon]
                    pPC = PatchCollection(aPatch, cmap=cmap, alpha=0.8, edgecolor=None, 
                                          facecolor=aColor, linewidths=0.25, 
                                          transform=cpl.crs.PlateCarree())
                    ax.add_collection(pPC)
                    if aExtent_in is None:
                        if pBoundary_in is None:
                            marginx = (dLongitude_max - dLongitude_min) / 20
                            marginy = (dLatitude_max - dLatitude_min) / 20
                            if dLongitude_min - marginx  < -180 or dLongitude_max + marginx > 180:
                                marginx = 0                       
                            if dLatitude_min - marginy < -90 or dLatitude_max + marginy > 90:
                                marginy = 0                       

                            aExtent = [dLongitude_min - marginx, dLongitude_max +
                                       marginx, dLatitude_min - marginy, dLatitude_max + marginy]
                        else:

                            marginx = (pBoundary_box[1] - pBoundary_box[0]) / 20
                            marginy = (pBoundary_box[3] - pBoundary_box[2]) / 20
                            aExtent = [pBoundary_box[0] - marginx, pBoundary_box[1] +
                                       marginx, pBoundary_box[2] - marginy, pBoundary_box[3] + marginy]
                    else:
                        aExtent = aExtent_in

                    ax.set_extent(aExtent)
                    ax.coastlines(color='black', linewidth=1)
                    ax.set_title(sTitle)
                    ax_cb = fig.add_axes([0.75, 0.15, 0.02, 0.6])

                    if iFlag_scientific_notation_colorbar == 1:
                        formatter = OOMFormatter(fformat="%1.1e")
                        cb = mpl.colorbar.ColorbarBase(ax_cb, orientation='vertical',
                                                       cmap=cmap,
                                                       norm=mpl.colors.Normalize(
                                                           dValue_min, dValue_max),  # vmax and vmin
                                                       extend=sExtend, format=formatter)
                    else:
                        formatter = OOMFormatter(fformat="%1.2f")
                        cb = mpl.colorbar.ColorbarBase(ax_cb, orientation='vertical',
                                                       cmap=cmap,
                                                       norm=mpl.colors.Normalize(
                                                           dValue_min, dValue_max),  # vmax and vmin
                                                       extend=sExtend, format=formatter)

                    cb.ax.get_yaxis().set_ticks_position('right')
                    cb.ax.get_yaxis().labelpad = 5
                    cb.ax.set_ylabel(sUnit, rotation=90, fontsize=iFont_size-2)
                    cb.ax.get_yaxis().set_label_position('left')
                    cb.ax.tick_params(labelsize=iFont_size-2)

                    gl = ax.gridlines(crs=cpl.crs.PlateCarree(), draw_labels=True,
                                      linewidth=1, color='gray', alpha=0.5, linestyle='--')
                    gl.xformatter = LONGITUDE_FORMATTER
                    gl.yformatter = LATITUDE_FORMATTER

                    gl.xlabel_style = {'size': 10, 'color': 'k',
                                       'rotation': 0, 'ha': 'right'}
                    gl.ylabel_style = {'size': 10, 'color': 'k',
                                       'rotation': 90, 'weight': 'normal'}



                    plt.savefig(sFilename_out, bbox_inches='tight')
                    plt.close('all')
                    plt.clf()
                    print(sFilename_out)
                    pass
                pass
            else:
                pass
            if iFlag_daily == 1:
                # Pre-calculate the color indices
                color_indices = np.clip(((aData - dValue_min) / (dValue_max - dValue_min) * 255).astype(int), 0, 255)
            

                for iStep in range(0, nTime):
                    sStep = '{:03d}'.format(iStep+1)
                    fig = plt.figure(dpi=iDPI)
                    fig.set_figwidth(iSize_x)
                    fig.set_figheight(iSize_y)
                    ax = fig.add_axes([0.08, 0.1, 0.62, 0.7],
                                      projection=pProjection_map)
                    ax.set_global()
                    # redefine the filename and title
                    sFilename_out = os.path.join(
                        sFolder_output_in, sVariable + '_' + sStep + '.png')

                    aPolygon = []
                    aColor = []
                    for i in range(0, nrow):
                        for j in range(0, ncolumn):
                            if not aMask[i, j]:
                                continue

                            dLongitude_cell_center = aLongitude_grid[i, j]
                            dLatitude_cell_center = aLatitude_grid[i, j]
                            dValue = aData[iStep, i, j]

                            # get its cell shape
                            aCoords_gcs = np.full((nVertex, 2), np.nan)

                            # Define coordinates for each vertex
                            coords = [(dLongitude_cell_center - half_res_x, dLatitude_cell_center - half_res_y),
                                      (dLongitude_cell_center + half_res_x, dLatitude_cell_center - half_res_y),
                                      (dLongitude_cell_center + half_res_x, dLatitude_cell_center + half_res_y),
                                      (dLongitude_cell_center - half_res_x, dLatitude_cell_center + half_res_y)]

                            # Fill aCoords_gcs using a loop
                            for idx, (x, y) in enumerate(coords):
                                aCoords_gcs[idx, :] = x, y

                            #dealing with missing value
                            if dValue > dValue_min:
                                iColor_index = color_indices[iStep, i, j]
                                # pick color from colormap       
                                aColor.append(cmap(iColor_index))
                                aPolygon.append(aCoords_gcs[:, 0:2])

                    # add extent
                    # Create a PatchCollection with the polygons
                    aPatch = [Polygon(poly, closed=True) for poly in aPolygon]
                    pPC = PatchCollection(aPatch, cmap=cmap, alpha=0.8, edgecolor=None, 
                                          facecolor=aColor, linewidths=0.25, 
                                          transform=cpl.crs.PlateCarree())
                    ax.add_collection(pPC)
                    if aExtent_in is None:
                        if pBoundary_in is None:
                            marginx = (dLongitude_max - dLongitude_min) / 20
                            marginy = (dLatitude_max - dLatitude_min) / 20
                            if dLongitude_min - marginx  < -180 or dLongitude_max + marginx > 180:
                                marginx = 0                       
                            if dLatitude_min - marginy < -90 or dLatitude_max + marginy > 90:
                                marginy = 0                       

                            aExtent = [dLongitude_min - marginx, dLongitude_max +
                                       marginx, dLatitude_min - marginy, dLatitude_max + marginy]
                        else:

                            marginx = (pBoundary_box[1] - pBoundary_box[0]) / 20
                            marginy = (pBoundary_box[3] - pBoundary_box[2]) / 20
                            aExtent = [pBoundary_box[0] - marginx, pBoundary_box[1] +
                                       marginx, pBoundary_box[2] - marginy, pBoundary_box[3] + marginy]
                    else:
                        aExtent = aExtent_in

                    ax.set_extent(aExtent)
                    ax.coastlines(color='black', linewidth=1)
                    ax.set_title(sTitle)
                    ax_cb = fig.add_axes([0.75, 0.15, 0.02, 0.6])

                    if iFlag_scientific_notation_colorbar == 1:
                        formatter = OOMFormatter(fformat="%1.1e")
                        cb = mpl.colorbar.ColorbarBase(ax_cb, orientation='vertical',
                                                       cmap=cmap,
                                                       norm=mpl.colors.Normalize(
                                                           dValue_min, dValue_max),  # vmax and vmin
                                                       extend=sExtend, format=formatter)
                    else:
                        formatter = OOMFormatter(fformat="%1.2f")
                        cb = mpl.colorbar.ColorbarBase(ax_cb, orientation='vertical',
                                                       cmap=cmap,
                                                       norm=mpl.colors.Normalize(
                                                           dValue_min, dValue_max),  # vmax and vmin
                                                       extend=sExtend, format=formatter)

                    cb.ax.get_yaxis().set_ticks_position('right')
                    cb.ax.get_yaxis().labelpad = 5
                    cb.ax.set_ylabel(sUnit, rotation=90, fontsize=iFont_size-2)
                    cb.ax.get_yaxis().set_label_position('left')
                    cb.ax.tick_params(labelsize=iFont_size-2)

                    gl = ax.gridlines(crs=cpl.crs.PlateCarree(), draw_labels=True,
                                      linewidth=1, color='gray', alpha=0.5, linestyle='--')
                    gl.xformatter = LONGITUDE_FORMATTER
                    gl.yformatter = LATITUDE_FORMATTER

                    gl.xlabel_style = {'size': 10, 'color': 'k',
                                       'rotation': 0, 'ha': 'right'}
                    gl.ylabel_style = {'size': 10, 'color': 'k',
                                       'rotation': 90, 'weight': 'normal'}



                    plt.savefig(sFilename_out, bbox_inches='tight')
                    plt.close('all')
                    plt.clf()
                    print(sFilename_out)

        print("Extraction successful!")
    except Exception as e:
        print(f"An error occurred: {e}")





import os

# normal job, no checkpoint


def slurm_prepare_job_script_python(iStart,
                                    iEnd,
                                    sBasename_checkpoint,
                                    sBasename_job,
                                    sBasename_python,
                                    sDirectory_job,
                                    sDirectory_python,
                                    sJob_name, iFlag_resubmit=None,
                                    iWalltime_in=None,
                                    nNode_in=None,
                                    nTask_in=None,
                                    sAccount=None,
                                    sEmail=None,
                                    sQueue_in=None):
    """
    Prepare a slurm job script for python with checkpoint

    Args:
        iStart (int): Checkpoint start index
        iEnd (int): Checkpoint end index
        sBasename_checkpoint (string): Basename 
        sBasename_job (string): Job name
        sBasename_python (string): Python path
        sDirectory_job (string): The path where contains the bash script
        sDirectory_python (string): The folder of Python
        sJob_name (string): Job name
        iFlag_resubmit (int, optional): Flag to indicate this is a resubmit. Defaults to None.
        iWalltime_in (int, optional): The walltime in hour. Defaults to None.
        nNode_in (int, optional): Number of node. Defaults to None.
        nTask_in (int, optional): Number of task. Defaults to None.
        sAccount (string, optional): Cluster account. Defaults to None.
        sEmail (string, optional): Email address. Defaults to None.
        sQueue_in (string, optional): Cluster partition. Defaults to None.
    """

    if nNode_in is not None:
        iNode = nNode_in
    else:
        iNode = 1

    if nTask_in is not None:
        nTask = nTask_in
    else:
        nTask = 10

    if sQueue_in is not None:
        sQueue = sQueue_in
    else:
        sQueue = 'short'

    if iWalltime_in is not None:
        iWalltime = iWalltime_in
    else:
        iWalltime = 2

    sStart = "{:0d}".format(iStart)
    sEnd = "{:0d}".format(iEnd)
    sNode = "{:0d}".format(iNode)
    sTask = "{:0d}".format(nTask)
    sWalltime = "{:0d}".format(iWalltime)
    os.chdir(sDirectory_job)

    pFile = open(sBasename_job, "w")  # write mode
    sLine = '#!/bin/bash' + '\n'
    pFile.write(sLine)

    if sAccount is not None:
        sLine = '#SBATCH --account=' + sAccount + '\n'
        pFile.write(sLine)

    sLine = '#SBATCH --begin=now+1minutes' + '\n'
    pFile.write(sLine)

    sLine = '#SBATCH --cpus-per-task=1 ' + '\n'
    pFile.write(sLine)

    sLine = '#SBATCH --dependency=singleton ' + '\n'
    pFile.write(sLine)
    sLine = '#SBATCH --error=stderr_%j.err' + '\n'
    pFile.write(sLine)

    if sJob_name is not None:
        sLine = '#SBATCH --job-name=' + sJob_name + \
            '  # create a name for your job' + '\n'
        pFile.write(sLine)

    sLine = '#SBATCH --mail-type=ALL' + '\n'
    pFile.write(sLine)

    if sEmail is not None:
        sLine = '#SBATCH --mail-user=' + sEmail + '\n'
        pFile.write(sLine)

    sLine = '#SBATCH --nodes=' + sNode + ' # node count' + '\n'
    pFile.write(sLine)
    sLine = '#SBATCH --ntasks=' + sTask + ' # total number of tasks' + '\n'
    pFile.write(sLine)

    sLine = '#SBATCH --output=stdout_%j.out' + '\n'
    pFile.write(sLine)

    sLine = '#SBATCH --partition=' + sQueue + '\n'  # can be improved here
    pFile.write(sLine)

    sLine = '#SBATCH --time=' + sWalltime + \
        ':00:00      # total run time limit (HH:MM:SS)' + '\n'
    pFile.write(sLine)

    sLine = 'module purge' + '\n'
    pFile.write(sLine)
    sLine = 'module load anaconda3/2019.03' + '\n'
    pFile.write(sLine)
    # only need to manuually source if the cluster requires
    # sLine = 'source /share/apps/anaconda3/2019.03/etc/profile.d/conda.sh' + '\n'
    # pFile.write( sLine )
    sLine = 'unset PYTHONHOME' + '\n'
    pFile.write(sLine)
    sLine = 'conda activate mpienv' + '\n'
    pFile.write(sLine)
    sLine = 'sDirectory_job=' + sDirectory_job + '\n'
    pFile.write(sLine)
    sLine = 'sDirectory_python=' + sDirectory_python + '\n'
    pFile.write(sLine)
    sLine = 'sBasename_python=' + sBasename_python + '\n'
    pFile.write(sLine)
    sLine = 'sFilename_python=' + '$sDirectory_python' + \
        "/" + '$sBasename_python' + '\n'
    pFile.write(sLine)

    sLine = 'echo " Job " ' + '${SLURM_JOBID}' + ' is launched' + '\n'
    pFile.write(sLine)

    sLine = 'conda deactivate' + '\n'
    pFile.write(sLine)

    if iFlag_resubmit == 1:  # advance job for checkpoint
        # now prepare the resubmit part
        sLine = 'iIndex="$(sed -n ' + ' 1p ' + \
            sBasename_checkpoint + ' | xargs)"' + '\n'
        pFile.write(sLine)
        sLine = 'iIndex="$(sed -n ' + ' 2p ' + \
            sBasename_checkpoint + ' | xargs)"' + '\n'
        pFile.write(sLine)
        sLine = 'iIndex="$(sed -n ' + ' 3p ' + \
            sBasename_checkpoint + ' | xargs)"' + '\n'
        pFile.write(sLine)

        sLine = 'if (($iIndex == 0));then' + '\n'
        pFile.write(sLine)
        sLine = '    echo "All tasks are finished"' + '\n'
        pFile.write(sLine)
        sLine = 'else' + '\n'
        pFile.write(sLine)
        sLine = '    echo "Job $SLURM_JOBID at $(date) will be re-submitted with new indices:, $sStart, $sEnd"' + '\n'
        pFile.write(sLine)
        sLine = '    sBasename_job="resubmit_${iIndex}.job"' + '\n'
        pFile.write(sLine)

        sLine = '    sbatch $sDirectory_job/$sBasename_job' + '\n'
        pFile.write(sLine)
        sLine = 'fi' + '\n'
        pFile.write(sLine)

    else:
        pass

    sLine = 'echo "Finished"' + '\n'
    pFile.write(sLine)
    pFile.close()

    return



import numpy as np
def remap( x, oMin, oMax, nMin, nMax ):

    #range check
    if oMin == oMax:
        print ("Warning: Zero input range")
        return None
    if nMin == nMax:
        print ("Warning: Zero output range")
        return None
    #check reversed input range
    reverseInput = False
    oldMin = np.min( np.array([oMin, oMax]) )
    oldMax = np.max( np.array([oMin, oMax]) )
    if not oldMin == oMin:
        reverseInput = True
    #check reversed output range
    reverseOutput = False   
    newMin = np.min( np.array([nMin, nMax]) )
    newMax = np.max( np.array([nMin, nMax]) )
    if not newMin == nMin :
        reverseOutput = True
    portion = (x-oldMin)*(newMax-newMin)/(oldMax-oldMin)
    if reverseInput:
        portion = (oldMax-x)*(newMax-newMin)/(oldMax-oldMin)
    result = portion + newMin
    if reverseOutput:
        result = newMax - portion
    return result


def convert_360_to_180(dLongitude_degree_in):
    """
   
    This function is modified from
    http://www.idlcoyote.com/map_tips/lonconvert.html
    

    Args:
        dLongitude_degree_in (float): The input longitude range from 0 to 360

    Returns:
        float: Longitude from -180 to 180
    """
    a = int(dLongitude_degree_in /180)
    dLongitude_out = dLongitude_degree_in - a*360.0

    return dLongitude_out



def convert_180_to_360(dLongitude_degree_in):
    """
   
    This function is modified from
    http://www.idlcoyote.com/map_tips/lonconvert.html
    

    Args:
        dLongitude_degree_in (float): The input longitude range from -180 to 180

    Returns:
        float: Longitude from 0 to 360
    """
    dLongitude_out = (dLongitude_degree_in + 360.0) % 360.0

    return dLongitude_out


import os
from osgeo import ogr, osr, gdal
def convert_vector_to_geojson(sFilename_vector_in, sFilename_geojson_out):
    #check input file
    if not os.path.exists(sFilename_vector_in):
        print('Input file not found')
        return
    #check if the file exists
    if os.path.exists(sFilename_geojson_out):
        os.remove(sFilename_geojson_out)
        pass

    #open the vector file
    pDataset_in = ogr.Open(sFilename_vector_in, 0)
    if pDataset_in is None:
        print('Could not open file')
        return
    #create the driver
    pDriver = ogr.GetDriverByName('GeoJSON')
    if pDriver is None:
        print('Driver not available')
        return
    #create the output file
    pDataset_out = pDriver.CreateDataSource(sFilename_geojson_out)
    if pDataset_out is None:
        print('Could not create file')
        return
    

    pSrs_out = osr.SpatialReference()
    pSrs_out.ImportFromEPSG(4326)    # WGS84 lat/lon
    wkt2 = pSrs_out.ExportToWkt()
    print(wkt2)

    
    #get the layer
    pLayer_in = pDataset_in.GetLayer()
    #obtain the geotype 
    iGeomType = pLayer_in.GetGeomType()
    #get the spatial reference
    pSrs_in = pLayer_in.GetSpatialRef()
    wkt1 = pSrs_in.ExportToWkt()
    print(wkt1)

    comparison = pSrs_in.IsSame(pSrs_out)    
    #if(comparison != 1):
    if( wkt1 != wkt2):
        iFlag_transform = 1
        transform = osr.CoordinateTransformation(pSrs_in, pSrs_out)
        pass
    else:
        iFlag_transform = 0
        pass

    if iGeomType == ogr.wkbPoint:
        #create the layer
        pLayer_out = pDataset_out.CreateLayer('layer', pSrs_out, geom_type=ogr.wkbPoint)
    else:
        if iGeomType == ogr.wkbLineString:
            #create the layer
            pLayer_out = pDataset_out.CreateLayer('layer', pSrs_out, geom_type=ogr.wkbLineString)
        else:
            if iGeomType == ogr.wkbPolygon:
                #create the layer
                pLayer_out = pDataset_out.CreateLayer('layer', pSrs_out, geom_type=ogr.wkbPolygon)
            else:
                print('Geometry type not supported')
                return
            pass
        pass
        
    if iFlag_transform == 1:
        
        #get the feature count 
        iFeatureCount = pLayer_in.GetFeatureCount()

        for i in range(iFeatureCount):
            pFeature_in = pLayer_in.GetFeature(i)    
            pGeometry = pFeature_in.GetGeometryRef()
            pGeometry.Transform(transform)
            pFeatureDefn = pLayer_in.GetLayerDefn()
            pFeature_out = ogr.Feature(pFeatureDefn)
            pFeature_out.SetGeometry(pGeometry)
            pLayer_out.CreateFeature(pFeature_out)            
        pass
    else:
        iFeatureCount = pLayer_in.GetFeatureCount()
        for i in range(iFeatureCount):
            pFeature_in = pLayer_in.GetFeature(i)    
            pGeometry = pFeature_in.GetGeometryRef()
            pFeatureDefn = pLayer_in.GetLayerDefn()
            pFeature_out = ogr.Feature(pFeatureDefn)
            pFeature_out.SetGeometry(pGeometry)
            pLayer_out.CreateFeature(pFeature_out)            
        pass

    pDataset_in = None
    pDataset_out = None
    
    print('Conversion completed')


    return



import os
import numpy as np
from osgeo import gdal, ogr, osr, gdalconst

def gdal_get_raster_spatial_ref_wkt(sFilename_in):
    ds = gdal.Open(sFilename_in, gdal.GA_ReadOnly)    
    pProjection =ds.GetProjection()
    return pProjection


import os
from osgeo import ogr, gdal

def convert_geojson_to_geoparquet(input_geojson, output_geoparquet):
    driver = ogr.GetDriverByName('Parquet')
    #check if the driver is available
    if driver is None:
        print('Parquet driver not available.')
        print("GDAL version:", gdal.__version__)
        return 
    
    driver2 = ogr.GetDriverByName('GeoJSON')
    in_dataset = driver2.Open(input_geojson, 0)
    in_layer = in_dataset.GetLayer()

    if os.path.exists(output_geoparquet):
        os.remove(output_geoparquet)

    out_dataset = driver.CreateDataSource(output_geoparquet)
    out_layer = out_dataset.CopyLayer(in_layer, 'flow')
    in_dataset = out_dataset = None

if __name__ == '__main__':
    input_geojson = '/compyfs/liao313/04model/pyhexwatershed/k34/pyhexwatershed20231001003/hexwatershed/00000001/variable_polygon.geojson'
    output_geoparquet = '/compyfs/liao313/04model/pyhexwatershed/k34/pyhexwatershed20231001003/hexwatershed/00000001/variable_polygon.parquet'
    input_geojson = '/compyfs/liao313/04model/pyhexwatershed/k34/pyhexwatershed20231001003/hexwatershed/00000001/flow_direction.geojson'
    output_geoparquet = '/compyfs/liao313/04model/pyhexwatershed/k34/pyhexwatershed20231001003/hexwatershed/00000001/flow_direction.parquet'
    input_geojson = '/compyfs/liao313/04model/pyhexwatershed/amazon/pyhexwatershed20240101012/pyflowline/dggrid.geojson'
    output_geoparquet = '/compyfs/liao313/04model/pyhexwatershed/amazon/pyhexwatershed20240101012/pyflowline/dggrid.parquet'
   
    convert_geojson_to_geoparquet(input_geojson, output_geoparquet)


import os
import xml.etree.ElementTree as ET
def parse_xml_file_atm(sFilename_xml_in):
    """
    Parse an XML file

    Args:
        sFilename_xml_in (string): The XML filename

    Returns:
        dict: The content in the XML file
    """

    if os.path.exists(sFilename_xml_in):
        pass
    else:
        print('The xml file does not exist!')
        return
    tree = ET.parse(sFilename_xml_in)
    root = tree.getroot()
    namelist = {}
    keys_str = []

    entry = root.findall('./fieldInfo/filePath')
    sFolder = (entry[0].text).strip()

    entry = root.findall('./fieldInfo/variableNames')
    dummy = (entry[0].text).strip()
    aField0 = dummy.split('\n')  #(dummy.split(" "))
    aField = list()
    for sField in aField0:
        d = sField.strip()
        e = d.split()
        aField.append(e[0])

    entry = root.findall('./fieldInfo/fileNames')
    dummy = (entry[0].text).strip()
    aFilename = dummy.split('\n')

    nf = len(aFilename)
    aFile=list()
    for i in range(nf):
        aFile.append( os.path.join( sFolder ,aFilename[i] ) )
       
    return sFolder, aField, aFile

def parse_xml_file_lnd(sFilename_xml_in):
    """
    Parse an XML file

    Args:
        sFilename_xml_in (string): The XML filename

    Returns:
        dict: The content in the XML file
    """

    if os.path.exists(sFilename_xml_in):
        pass
    else:
        print('The xml file does not exist!')
        return
    tree = ET.parse(sFilename_xml_in)
    root = tree.getroot()

    entry = root.findall('./domainInfo/filePath')
    sFolder_domain = (entry[0].text).strip()
    entry = root.findall('./domainInfo/variableNames')
    dummy = (entry[0].text).strip()
    aField0 = dummy.split('\n')  #(dummy.split(" "))
    aField_domain = list()
    for sField in aField0:
        d = sField.strip()
        e = d.split()
        aField_domain.append(e[0])
        pass

    entry = root.findall('./domainInfo/fileNames')
    dummy = (entry[0].text).strip()
    sFilename_domain = dummy.split('\n')
    sFilename_domain = os.path.join( sFolder_domain ,sFilename_domain[0] )
   

    entry = root.findall('./fieldInfo/filePath')
    sFolder = (entry[0].text).strip()

    entry = root.findall('./fieldInfo/variableNames')
    dummy = (entry[0].text).strip()
    aField0 = dummy.split('\n')  #(dummy.split(" "))
    aField = list()
    for sField in aField0:
        d = sField.strip()
        e = d.split()
        aField.append(e[0])

    entry = root.findall('./fieldInfo/fileNames')
    dummy = (entry[0].text).strip()
    aFilename = dummy.split('\n')

    nf = len(aFilename)
    aFile=list()
    for i in range(nf):
        aFile.append( os.path.join( sFolder ,aFilename[i] ) )
       
    return sFilename_domain,aField_domain, sFolder, aField, aFile


import os, sys
import numpy as np
from osgeo import gdal, osr

def gdal_read_geotiff_file(sFilename_in, iFlag_metadata_only = 0):
    """Read a Geotiff format raster file.

    Args:
        sFilename_in (string): The file name

    Returns:
        tuple: aData_out, dPixelWidth, dOriginX, dOriginY, nrow, ncolumn, dMissing_value , pGeotransform, pProjection,  pSpatial_reference
    """
    
    if os.path.exists(sFilename_in):
        pass
    else:
        print('The file does not exist!')
        return
    
    

    sDriverName='GTiff'
    pDriver = gdal.GetDriverByName(sDriverName)  

    if pDriver is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName)  

    pDataset = gdal.Open(sFilename_in, gdal.GA_ReadOnly)

    if pDataset is None:
        print("Couldn't open this file: " + sFilename_in)
        sys.exit("Try again!")
    else:       
        dMissing_value = -9999.0
        pProjection = pDataset.GetProjection()

        pDataset.GetMetadata()
       
        ncolumn = pDataset.RasterXSize
        nrow = pDataset.RasterYSize
        nband = pDataset.RasterCount

        pGeotransform = pDataset.GetGeoTransform()
        dOriginX = pGeotransform[0]
        dOriginY = pGeotransform[3]
        dPixelWidth = pGeotransform[1]
        dPixelHeight = pGeotransform[5]
        #we will use one of them to keep the consistency
        pSpatial_reference = osr.SpatialReference(wkt=pProjection)  

        if iFlag_metadata_only ==1:
            pass
        else:    
            if nband >1: 
                print('This is a multi-band raster file!')
                print('We will only read the first band!')
                pass

            pBand = pDataset.GetRasterBand(1)

            # Data type of the values
            
            iData_type = pBand.DataType
            sDate_type = gdal.GetDataTypeName(iData_type)
            # Compute statistics if needed
            if pBand.GetMinimum() is None or pBand.GetMaximum() is None:
                pBand.ComputeStatistics(0)

            dMissing_value = pBand.GetNoDataValue()
            aData_out = pBand.ReadAsArray(0, 0, ncolumn, nrow)                

        pDataset = None
        pBand = None             

        if iFlag_metadata_only == 1:
            pData = {
                'pixelWidth': dPixelWidth,
                'pixelHeight': dPixelHeight,
                'originX': dOriginX,
                'originY': dOriginY,
                'nrow': nrow,
                'ncolumn': ncolumn,
                #'missingValue': dMissing_value,
                'geotransform': pGeotransform,
                'projection': pProjection,
                'spatialReference': pSpatial_reference
            }
            return pData
           
        else:   
            pData = {
                'dataOut': aData_out,
                'dataType': iData_type, # 'Float32', 'Float64', 'UInt16', 'Int16', 'UInt32', 'Int32', 'Byte', 'UInt32', 'Int32', 'CInt16', 'CInt32', 'CFloat32', 'CFloat64
                'pixelWidth': dPixelWidth,
                'pixelHeight': dPixelHeight,
                'originX': dOriginX,
                'originY': dOriginY,
                'nrow': nrow,
                'ncolumn': ncolumn,
                'missingValue': dMissing_value,
                'geotransform': pGeotransform,
                'projection': pProjection,
                'spatialReference': pSpatial_reference
            }
            return pData

def gdal_read_geotiff_file_multiple_band(sFilename_in, iFlag_metadata_only = 0):
    
    """Read a Geotiff format raster file with multiple bands.

    Args:
        sFilename_in (string): The filename

    Returns:
        tuple: aData_out, dPixelWidth, dOriginX, dOriginY, nrow, ncolumn, dMissing_value , pGeotransform, pProjection,  pSpatial_reference
    """
    
    if os.path.exists(sFilename_in):
        pass
    else:
        print('The file does not exist!')
        return
   
    sDriverName='GTiff'
    pDriver = gdal.GetDriverByName(sDriverName)  

    if pDriver is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName) 
    pDataset = gdal.Open(sFilename_in, gdal.GA_ReadOnly)

    if pDataset is None:
        print("Couldn't open this file: " + sFilename_in)
      
        sys.exit("Try again!")
    else:        
        pProjection = pDataset.GetProjection()
        pDataset.GetMetadata()
       
        ncolumn = pDataset.RasterXSize
        nrow = pDataset.RasterYSize
        nband = pDataset.RasterCount
       
        pGeotransform = pDataset.GetGeoTransform()
        dOriginX = pGeotransform[0]
        dOriginY = pGeotransform[3]
        dPixelWidth = pGeotransform[1]
        dPixelHeight = pGeotransform[5]

        pBand = pDataset.GetRasterBand(1)
        
        dummy_data = pBand.ReadAsArray(0, 0, ncolumn, nrow)
        dt = type(dummy_data)
        #there is a chance that GDAL datetype is not compatiable with numpy datatype.
        dMissing_value = pBand.GetNoDataValue()
        pSpatial_reference = osr.SpatialReference(wkt=pProjection)

        if iFlag_metadata_only ==1:
            pass
        else:    
            aData_out = np.full( (nband, nrow, ncolumn) , -9999.0, dtype= dt )
            for iBand in range(nband):
                pBand = pDataset.GetRasterBand( iBand + 1)
                iData_type = pBand.DataType
                aData_out[iBand, :, :] = pBand.ReadAsArray(0, 0, ncolumn, nrow)
                pass

        pDriver = None
        pDataset = None
        pBand = None
        if iFlag_metadata_only == 1:
            pData = {
                'pixelWidth': dPixelWidth,
                'pixelHeight': dPixelHeight,
                'originX': dOriginX,
                'originY': dOriginY,
                'nrow': nrow,
                'ncolumn': ncolumn,
                #'missingValue': dMissing_value,
                'geotransform': pGeotransform,
                'projection': pProjection,
                'spatialReference': pSpatial_reference
            }
            return pData
        else:
            pData = {
                'dataOut': aData_out,
                'dataType': iData_type,
                'pixelWidth': dPixelWidth,
                'pixelHeight': dPixelHeight,
                'originX': dOriginX,
                'originY': dOriginY,
                'nrow': nrow,
                'ncolumn': ncolumn,
                'missingValue': dMissing_value,
                'geotransform': pGeotransform,
                'projection': pProjection,
                'spatialReference': pSpatial_reference
            }
            return pData




import numpy as np
from pyearth.gis.location.convert_longitude_latitude_to_sphere_3d import convert_longitude_latitude_to_sphere_3d

def calculate_angle_between_vertex_normal(dLongitude1_in, dLatitude1_in, 
                                dLongitude2_in, dLatitude2_in, 
                                dLongitude3_in, dLatitude3_in, 
                                iFlag_radian = None):
    if iFlag_radian is None:
        dLongitude1_radian_in, dLatitude1_radian_in = np.radians(np.array((dLongitude1_in, dLatitude1_in) ))
        dLongitude2_radian_in, dLatitude2_radian_in = np.radians(np.array((dLongitude2_in, dLatitude2_in) )) #this is the middle one
        dLongitude3_radian_in, dLatitude3_radian_in = np.radians(np.array((dLongitude3_in, dLatitude3_in) ))
        pass
    else:         
        dLongitude1_radian_in, dLatitude1_radian_in = dLongitude1_in , dLatitude1_in  
        dLongitude2_radian_in, dLatitude2_radian_in = dLongitude2_in , dLatitude2_in #this is the middle one
        dLongitude3_radian_in, dLatitude3_radian_in = dLongitude3_in , dLatitude3_in
        pass
    # The points in 3D space
    a3 = convert_longitude_latitude_to_sphere_3d(dLongitude1_radian_in, dLatitude1_radian_in)
    b3 = convert_longitude_latitude_to_sphere_3d(dLongitude2_radian_in, dLatitude2_radian_in)
    c3 = convert_longitude_latitude_to_sphere_3d(dLongitude3_radian_in, dLatitude3_radian_in)
    a3vec = a3 - b3
    c3vec = c3 - b3 
    dot = np.dot(a3vec, c3vec)
    g = np.cross(a3vec, c3vec)
    det = np.dot(b3, g)
    angle = np.arctan2(det, dot)
    f = np.degrees(angle) 
    if f < 0:
        f = 360 + f
    
    return f


if __name__ == '__main__':

    dLongitude1_in=-149
    dLatitude1_in= 71
 
    dLongitude2_in=-148.1875
    dLatitude2_in=70.125
    
    dLongitude3_in=-147
    dLatitude3_in= 70.125
                          
    

    dAngle = calculate_angle_between_vertex_normal(dLongitude1_in, dLatitude1_in, 
                                dLongitude2_in, dLatitude2_in, 
                                dLongitude3_in, dLatitude3_in 
                               )
    print(dAngle)


    #calculate left or right of line segment

    #a
    dLongitude1_in=-149
    dLatitude1_in=70

    #b
    dLongitude2_in=-148
    dLatitude2_in= 70   
    
    #c
    dLongitude3_in=-147
    dLatitude3_in= 70

    #d
    dLongitude4_in=-149
    dLatitude4_in= 72

    dLongitude1_radian_in, dLatitude1_radian_in = np.radians(np.array((dLongitude1_in, dLatitude1_in) ))
    dLongitude2_radian_in, dLatitude2_radian_in = np.radians(np.array((dLongitude2_in, dLatitude2_in) )) #this is the middle one
    dLongitude3_radian_in, dLatitude3_radian_in = np.radians(np.array((dLongitude3_in, dLatitude3_in) ))
    dLongitude4_radian_in, dLatitude4_radian_in = np.radians(np.array((dLongitude4_in, dLatitude4_in) ))
    a3 = convert_longitude_latitude_to_sphere_3d(dLongitude1_radian_in, dLatitude1_radian_in)
    b3 = convert_longitude_latitude_to_sphere_3d(dLongitude2_radian_in, dLatitude2_radian_in)
    c3 = convert_longitude_latitude_to_sphere_3d(dLongitude3_radian_in, dLatitude3_radian_in)  
    d3 = convert_longitude_latitude_to_sphere_3d(dLongitude4_radian_in, dLatitude4_radian_in)

    a3vec = b3 - a3
    c3vec = c3 - b3 
    d3vec = b3 - d3


    dot = np.dot(c3vec, a3vec)
    g = np.cross(c3vec, a3vec)
    det = np.dot(b3, g)
    angle1 = np.arctan2(det, dot)
    f1 = np.degrees(angle1) 
    if f1 < 0:
        f1 = 360 + f1
    

    dot = np.dot(d3vec, a3vec)
    g = np.cross(d3vec, a3vec)
    det = np.dot(b3, g)
    angle2 = np.arctan2(det, dot)
    f2 = np.degrees(angle2) 
    if f2 < 0:
        f2 = 360 + f2
                          
    
    print(f1)
    print(f2)

    delta = f1 - f2 
    print(delta)
    


import numpy as np
from pyearth.system.define_global_variables import *

def calculate_distance_based_on_longitude_latitude(dLongitude_from,
                                                   dLatitude_from,
                                                   dLongitude_to,
                                                   dLatitude_to,
                                                   iFlag_radian = None,
                                                   dRadius_in = None):
    """
    Calculate the great circle distance between two points 
    on the earth (specified in decimal degrees)
    https://en.wikipedia.org/wiki/Great-circle_distance
    The current algorothm is the  haversine formula 

    Args:
        dLongitude_radian_from (float): The longitude of the start point
        dLatitude_radian_from (float):  The latitude of the start point
        dLongitude_radian_to (float):  The longitude of the end point
        dLatitude_radian_to (float):  The latitude of the end point

    Returns:
        float: The great circle distance
    """
    
    if iFlag_radian is None:# convert decimal degrees to radians
        dLongitude_radian_from, dLatitude_radian_from, dLongitude_radian_to, dLatitude_radian_to = np.radians([dLongitude_from,
                                                                                                          dLatitude_from,
                                                                                                          dLongitude_to,
                                                                                                          dLatitude_to])
    else: #already in radian
        dLongitude_radian_from, dLatitude_radian_from, dLongitude_radian_to, dLatitude_radian_to = dLongitude_from, dLatitude_from, dLongitude_to, dLatitude_to
        pass

    # haversine formula
    dlon = dLongitude_radian_to - dLongitude_radian_from
    dlat = dLatitude_radian_to - dLatitude_radian_from

    a = np.sin(dlat/2)**2 + np.cos(dLatitude_radian_from) * np.cos(dLatitude_radian_to) * np.sin(dlon/2)**2
    
    
    c = 2 * np.arcsin(np.sqrt(a))

    if iFlag_radian is not None:
        return c
            
    else:  
        if dRadius_in is not None:
            d = c * dRadius_in
        else:
            d = c * earth_radius
        
        return d    

    




import osgeo
from osgeo import  osr
def reproject_coordinates(dx_in, dy_in, pSpatial_reference_source_in, pSpatial_reference_target_in=None):
    """ Reproject a pair of x,y coordinates. 

    Args:
        dx_in (float): X Coordinate of point
        dy_in (float): Y Coordinate of
        pSpatial_reference_source_in (osr): The source spatial reference of point
        pSpatial_reference_target_in (osr, optional): The target spatial reference of point. Defaults to None.

    Returns:
        Tuple: dx_new, dy_new
    """

    if pSpatial_reference_target_in is not None:

        pass
    else:
        pSpatial_reference_target_in = osr.SpatialReference()
        pSpatial_reference_target_in.ImportFromEPSG(4326)
        
        pass

    
    if int(osgeo.__version__[0]) >= 3:
    # GDAL 3 changes axis order: https://github.com/OSGeo/gdal/issues/1546
                    
        pSpatial_reference_source_in.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)
        pSpatial_reference_target_in.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)

    
    pTransform = osr.CoordinateTransformation( pSpatial_reference_source_in, pSpatial_reference_target_in)
   
    dx_out, dy_out, z = pTransform.TransformPoint( dx_in,dy_in)
    
    return dx_out, dy_out

def reproject_coordinates_batch(aX_in, aY_in, pSpatial_reference_source_in, pSpatial_reference_target_in=None):
    """ Reproject a list of x, y coordinates.

    Args:
        aX_in (list): A list of X Coordinate of points
        aY_in (list): A list of Y Coordinate of points
        pSpatial_reference_source_in (osr): The source spatial reference of point
        pSpatial_reference_target_in (osr, optional): The target spatial reference of point. Defaults to None.

    Returns:
        Tuple: aX_out, aY_out
    """

    if pSpatial_reference_target_in is not None:

        pass
    else:
        pSpatial_reference_target_in = osr.SpatialReference()
        pSpatial_reference_target_in.ImportFromEPSG(4326)
        
        pass

    
    if int(osgeo.__version__[0]) >= 3:
    # GDAL 3 changes axis order: https://github.com/OSGeo/gdal/issues/1546
                    
        pSpatial_reference_source_in.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)
        pSpatial_reference_target_in.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)

    
    pTransform = osr.CoordinateTransformation( pSpatial_reference_source_in, pSpatial_reference_target_in)

    npoint = len(aX_in)
    aX_out=list()
    aY_out=list()
    for i in range(npoint):
        x0 = aX_in[i]
        y0 = aY_in[i]
   
        x1,y1, z = pTransform.TransformPoint( x0,y0)

        aX_out.append(x1)
        aY_out.append(y1)
    
    return aX_out,aY_out



import os, sys
import numpy as np
from osgeo import gdal, osr


def gdal_read_envi_file(sFilename_in):
    """Read a ENVI standard format raster file.

    Args:
        sFilename_in (string): The filename

    Returns:
        tuple: aData_out, pPixelWidth, dOriginX, dOriginY, nrow, ncolumn, dMissing_value , pGeotransform, pProjection,  pSpatial_reference
    """
    
    if os.path.exists(sFilename_in):
        pass
    else:
        print('The file does not exist!')
        return

    sDriverName='ENVI'
    pDriver = gdal.GetDriverByName(sDriverName)  

    if pDriver is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName)
         
    pDataset = gdal.Open(sFilename_in, gdal.GA_ReadOnly)

    if pDataset is None:
        print("Couldn't open this file: " + sFilename_in)
        print('Perhaps you need an ENVI .hdr file?')
        
        sys.exit("Try again!")
    else:
        
        pProjection = pDataset.GetProjection()
       
        ncolumn = pDataset.RasterXSize
        nrow = pDataset.RasterYSize
        nband = pDataset.RasterCount

        pGeotransform = pDataset.GetGeoTransform()
        dOriginX = pGeotransform[0]
        dOriginY = pGeotransform[3]
        pPixelWidth = pGeotransform[1]
        pPixelHeight = pGeotransform[5]
       
        pBand = pDataset.GetRasterBand(1)
        dMissing_value = pBand.GetNoDataValue()
        aData_out = pBand.ReadAsArray(0, 0, ncolumn, nrow)
        pSpatial_reference = osr.SpatialReference(wkt=pProjection)

        pDriver = None
        pDataset = None
        pBand = None

        return aData_out, pPixelWidth, dOriginX, dOriginY, nrow, ncolumn, dMissing_value , pGeotransform, pProjection,  pSpatial_reference


def gdal_read_envi_file_multiple_band(sFilename_in):
    """Read a ENVI standard format raster file with multiple bands.

    Args:
        sFilename_in (string): The file name

    Returns:
        tuple: aData_out, pPixelWidth, dOriginX, dOriginY, nband, nrow, ncolumn, dMissing_value , pGeotransform, pProjection, pSpatial_reference
    """
    if os.path.exists(sFilename_in):
        pass
    else:
        print('The file does not exist!')
        return

    sDriverName='ENVI'
    pDriver = gdal.GetDriverByName(sDriverName)  

    if pDriver is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName) 

    pDataset = gdal.Open(sFilename_in, gdal.GA_ReadOnly)

    if pDataset is None:
        print("Couldn't open this file: " + sFilename_in)
        print('Perhaps you need an ENVI .hdr file?')
        
        sys.exit("Try again!")
    else:
        
        pProjection = pDataset.GetProjection()
       
        ncolumn = pDataset.RasterXSize
        nrow = pDataset.RasterYSize
        nband = pDataset.RasterCount

        pGeotransform = pDataset.GetGeoTransform()
        dOriginX = pGeotransform[0]
        dOriginY = pGeotransform[3]
        pPixelWidth = pGeotransform[1]
        pPixelHeight = pGeotransform[5]

        pBand = pDataset.GetRasterBand(1)
        dt = gdal.GetDataTypeName(pBand.DataType)
        dMissing_value = pBand.GetNoDataValue()
        #there is a chance that GDAL datetype is not compatiable with numpy datatype.

        aData_out = np.full( (nband, nrow, ncolumn) , -9999.0, dtype= np.float32 )
        for iBand in range(nband):
            pBand = pDataset.GetRasterBand( iBand + 1)
            
            aData_out[iBand, :, :] = pBand.ReadAsArray(0, 0, ncolumn, nrow)

        pSpatial_reference = osr.SpatialReference(wkt=pProjection)

        pDriver = None
        pDataset = None
        pBand = None

        return aData_out, pPixelWidth, dOriginX, dOriginY, nband, nrow, ncolumn, dMissing_value , pGeotransform, pProjection, pSpatial_reference


from osgeo import ogr, osr, gdal
def get_utm_spatial_reference(dLongitude_in):
    if -180 <= dLongitude_in <= 180:
        zone = int((dLongitude_in + 180) / 6) + 1
        hemisphere = 'N' if dLongitude_in >= 0 else 'S'
        epsg_code = 32600 + zone if hemisphere == 'N' else 32700 + zone
        utm_sr = osr.SpatialReference()
        utm_sr.ImportFromEPSG(epsg_code)
        return utm_sr
    else:
        raise ValueError("Longitude must be in the range [-180, 180].")


import os
from osgeo import gdal
def gdal_write_envi_file(sFilename_in,
                         aData_in,
                         dPixelWidth_in, dOriginX_in,
                         dOriginY_in, dMissing_value_in,
                         pSpatial_reference_in):
    """
    Write a ENVI standard format raster file

    Args:
        sFilename_in (string): The filename
        aData_in (numpy.array): The data
        dPixelWidth_in (float): The resolution
        dOriginX_in (float): The location of origin x
        dOriginY_in (float): The location of origin y
        dMissing_value_in (float): The missinge value
        pSpatial_reference_in (osr): The spatial reference

    Returns:
        Tuple: pGeotransform_out, pProjection_out
    """
    if os.path.exists(sFilename_in):
        os.remove(sFilename_in)
        pass

    sDriverName = 'ENVI'
    pDriver = gdal.GetDriverByName(sDriverName)

    if pDriver is None:
        print("%s pDriver not available.\n" % sDriverName)
    else:
        print("%s pDriver IS available.\n" % sDriverName)

    nrow, ncolumn = aData_in.shape
    nband = 1

    # Creates a new raster data source
    pDataset = pDriver.Create(sFilename_in,
                              ncolumn,
                              nrow,
                              nband,
                              gdal.GDT_Float32)

    # Write metadata

    pDataset.SetGeoTransform([dOriginX_in,
                              dPixelWidth_in,
                              0.0,
                              dOriginY_in,
                              0.0,
                              -dPixelWidth_in])

    pProjection = pSpatial_reference_in.ExportToPrettyWkt()
    pDataset.SetProjection(pProjection)

    # Write raster datasets
    pBand = pDataset.GetRasterBand(1)
    pBand.WriteArray(aData_in)
    pBand.SetNoDataValue(dMissing_value_in)

    pGeotransform_out = pDataset.GetGeoTransform()
    pProjection_out = pDataset.GetProjection()

    pDataset.FlushCache()  # Write to disk.
    pDriver = None
    pDataset = None
    pBand = None

    return pGeotransform_out, pProjection_out


def gdal_write_envi_file_multiple_band(sFilename_in,
                                       aData_in,
                                       dPixelWidth_in,
                                       dOriginX_in,
                                       dOriginY_in,
                                       dMissing_value_in,
                                       pSpatial_reference_in):
    """
    Write a multi-band ENVI raster file

    Args:
        sFilename_in (string): The filename
        aData_in (numpy.array): The data
        dPixelWidth_in (float): The resolution
        dOriginX_in (float): The location of origin x
        dOriginY_in (float): The location of origin y
        dMissing_value_in (float): The missinge value
        pSpatial_reference_in (osr): The spatial reference

    Returns:
        Tuple: pGeotransform_out, pProjection_out
    """

    if os.path.exists(sFilename_in):
        os.remove(sFilename_in)
        pass

    sDriverName = 'ENVI'
    pDriver = gdal.GetDriverByName(sDriverName)

    if pDriver is None:
        print("%s pDriver not available.\n" % sDriverName)
    else:
        print("%s pDriver IS available.\n" % sDriverName)

    nband, nrow, ncolumn = aData_in.shape

    # Creates a new raster data source
    pDataset = pDriver.Create(sFilename_in,
                              ncolumn,
                              nrow,
                              nband,
                              gdal.GDT_Float32)

    # Write metadata

    pDataset.SetGeoTransform([dOriginX_in,
                              dPixelWidth_in,
                              0.0,
                              dOriginY_in,
                              0.0,
                              -dPixelWidth_in])

    pProjection = pSpatial_reference_in.ExportToPrettyWkt()
    pDataset.SetProjection(pProjection)

    # Write raster datasets
    for iBand in range(nband):
        pBand = pDataset.GetRasterBand(iBand + 1)
        pBand.WriteArray(aData_in[iBand, :, :])
        pBand.SetNoDataValue(dMissing_value_in)

    pGeotransform_out = pDataset.GetGeoTransform()
    pProjection_out = pDataset.GetProjection()

    pDataset.FlushCache()  # Write to disk.
    pDriver = None
    pDataset = None
    pBand = None

    return sFilename_in, pGeotransform_out, pProjection_out



import copy
import numpy as np
def remove_outliers(aData_in, dOutlier_percentage):
    """
    Remove the outliers within the datesets
    Args:
        aData_in (numpy.array): The array data
        dOutlier_percentage (float): The percentage to be marked as outlier

    Returns:
        numpy.array: The data without outlier
    """
    
    #remove nan first    
    a = copy.deepcopy(aData_in) 
    b = np.array(a)
    c = b.ravel() #flatted array
    good_index =np.where( np.isfinite(c) == True  )
    d = c[good_index]
    upper_quartile = np.percentile(d, 95)
    lower_quartile = np.percentile(d, 5)
    dQR = (upper_quartile - lower_quartile) * dOutlier_percentage
    quartileSet = (lower_quartile + dQR, upper_quartile - dQR)    

    dummy_index = np.where( (b>=quartileSet[0]) & (b<=quartileSet[1]) )

    return aData_in[dummy_index]


import os
from osgeo import osr, ogr

def reproject_vector(sFilename_vector_in, sFilename_vector_out, pProjection_target):

    pDataset_vector = ogr.Open(sFilename_vector_in)    
    # Get the first layer in the shapefile
    pLayer_vector = pDataset_vector.GetLayer(0)   
    pSpatial_reference_vector = pLayer_vector.GetSpatialRef()
    pProjection_source = pSpatial_reference_vector.ExportToWkt()        

    pSpatial_reference_target = osr.SpatialReference()
    pSpatial_reference_target.ImportFromWkt(pProjection_target)

    pProjection_target = pSpatial_reference_target.ExportToWkt()

    #get the extension of polygon file  
    sExtension_vector = os.path.splitext(sFilename_vector_in)[1]
    #get the driver for the extension
    if sExtension_vector == '.geojson':
        pDriver_vector = ogr.GetDriverByName('GeoJSON')
    else:
        pDriver_vector = ogr.GetDriverByName(sExtension_vector)

    if (pProjection_source != pProjection_target):

        transform = osr.CoordinateTransformation(pSpatial_reference_vector, pSpatial_reference_target)

        pDataset_transform = pDriver_vector.CreateDataSource(sFilename_vector_out)
        #create a new shapefile layer
        pLayer_transform = pDataset_transform.CreateLayer('transform', pSpatial_reference_target, geom_type=ogr.wkbPolygon)
        #create a new feature
        pFeature_transform = ogr.Feature(pLayer_transform.GetLayerDefn())
        pFeature_clip = pLayer_vector.GetNextFeature()
        pPolygon = pFeature_clip.GetGeometryRef()
        pPolygon.Transform(transform)      
        #set the geometry
        pFeature_transform.SetGeometry(pPolygon)
        #add the feature to the layer
        pLayer_transform.CreateFeature(pFeature_transform)        
        pDataset_transform.FlushCache()
        #use the new shapefile to clip the raster
    else:
        print('The input and target vector files have the same spatial reference')    

    pSpatial_reference_vector = None
    pSpatial_reference_target = None
    return


import os
import numpy as np
from osgeo import osr, gdal, ogr

def merge_vector_polygon_data(iFiletype_in,
                              aFilename_in,
                              aVariable_in,
                              sFilename_out,
                              sVariable_out,
                              dMissing_value_in=None,
                              dData_max_in=None,
                              dData_min_in=None,
                              ):
    """
    Map a vector polyline data

    Args:

    """
    sVar = sVariable_out[0:4].lower()

    if iFiletype_in == 1:  # geojson
        pDriver = ogr.GetDriverByName('GeoJSON')
    else:
        if iFiletype_in == 2:  # shapefile
            pDriver = ogr.GetDriverByName('Esri Shapefile')

    nDataset = len(aFilename_in)
    aDataset = []
    for sFilename_in in aFilename_in:
        pDataset = pDriver.Open(sFilename_in, gdal.GA_ReadOnly)
        aDataset.append(pDataset)

    pLayer = aDataset[0].GetLayer(0)

    pSpatial_reference_gcs = osr.SpatialReference()
    pSpatial_reference_gcs.ImportFromEPSG(4326)    # WGS84 lat/lon

    # get the number of features
    nFeature = pLayer.GetFeatureCount()

    if os.path.exists(sFilename_out):
        os.remove(sFilename_out)

    pDataset_out = pDriver.CreateDataSource(sFilename_out)
    pLayer_out = pDataset_out.CreateLayer(
        'cell', pSpatial_reference_gcs, ogr.wkbPolygon)
    # Add one attribute
    # long type for high resolution
    pLayer_out.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64))
    # long type for high resolution
    pLayer_out.CreateField(ogr.FieldDefn(sVar, ogr.OFTReal))
    pLayerDefn = pLayer_out.GetLayerDefn()
    pFeature_out = ogr.Feature(pLayerDefn)

    lID = 1
    for iFeature in range(nFeature):
        aValue = list()
        for iData in range(nDataset):
            pLayer = aDataset[iData].GetLayer(0)
            pFeature = pLayer.GetFeature(iFeature)
            if iData == 0:  # use the first data as geometry
                pGeometry_in = pFeature.GetGeometryRef()
                pGeometry_out = pGeometry_in.Clone()

            sVar_dummy = aVariable_in[iData].lower()
            sVar_dummy = sVar_dummy[0:4]
            dValue = float(pFeature.GetField(sVar_dummy))
            aValue.append(dValue)

        aValue = np.array(aValue)
        # take the sum
        dValue = np.sum(aValue)

        pFeature_out.SetGeometry(pGeometry_out)
        pFeature_out.SetField('cellid', lID)
        pFeature_out.SetField(sVar, float(dValue))
        pLayer_out.CreateFeature(pFeature_out)
        lID = lID + 1

    pDataset = pLayer = pFeature = None  # save, close

    print("finished")




from datetime import datetime
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from pyearth.system.define_global_variables import *
from pyearth.visual.color.create_diverge_rgb_color_hex import create_diverge_rgb_color_hex
from pyearth.visual.color.choose_n_color import polylinear_gradient, rand_hex_color


def plot_time_series_data_w_variation(aTime_all,
                                      aData_all,
                                      aData_upper_all,
                                      aData_lower_all,
                                      sFilename_out,
                                      iDPI_in=None,
                                      iFlag_log_in=None,
                                      iFlag_miniplot_in=None,
                                      iFlag_scientific_notation_in=None,
                                      ncolumn_in=None,
                                      aFlag_trend_in=None,
                                      iReverse_y_in=None,
                                      iSize_x_in=None,
                                      iSize_y_in=None,
                                      dMax_x_in=None,
                                      dMin_x_in=None,
                                      dMax_y_in=None,
                                      dMin_y_in=None,
                                      dSpace_y_in=None,
                                      aMarker_in=None,
                                      aColor_in=None,
                                      aLinestyle_in=None,
                                      sLabel_y_in=None,
                                      aLabel_legend_in=None,
                                      aLocation_legend_in=None,
                                      sDate_type_in=None,
                                      sFormat_y_in=None,
                                      sLocation_legend_in=None,
                                      sTitle_in=None):
    # find how many data will be plotted
    aTime_all = np.array(aTime_all)
    aData_all = np.array(aData_all)
    pShape = aData_all.shape

    nData = pShape[0]

    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_log_in is not None:
        iFlag_log = iFlag_log_in
    else:
        iFlag_log = 0

    if iFlag_scientific_notation_in is not None:
        iFlag_scientific_notation = iFlag_scientific_notation_in
    else:
        iFlag_scientific_notation = 0

    if aFlag_trend_in is not None:
        aFlag_trend = aFlag_trend_in
    else:
        aFlag_trend = np.full(nData, 0)

    if iReverse_y_in is not None:
        iReverse_y = iReverse_y_in
    else:
        iReverse_y = 0

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12

    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9

    if sLabel_y_in is not None:
        sLabel_y = sLabel_y_in
    else:
        sLabel_y = ''

    if aLabel_legend_in is not None:
        aLabel_legend = aLabel_legend_in
    else:
        aLabel_legend = np.full(nData, '')

    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    if aMarker_in is not None:
        aMarker = aMarker_in
    else:
        aMarker = np.full(nData, '+')

    if aColor_in is not None:
        aColor = aColor_in
        aColor_fill = aColor
    else:
        if (nData >= 3):
            if (nData <= 12):
                aColor = create_diverge_rgb_color_hex(nData)
                aColor_fill = aColor
            else:
                a = rand_hex_color(num=2)
                b = polylinear_gradient(a, nData)
                aColor = b['hex']
                aColor_fill = aColor
                pass
        else:
            if nData == 2:
                aColor = ['red', 'blue']
                aColor_fill = aColor
            else:
                aColor = ['navy']
                aColor_fill = ['lightblue']

    if aLinestyle_in is not None:
        aLinestyle = aLinestyle_in
    else:
        aLinestyle = np.full(nData, '-')

    if dMax_x_in is not None:
        dMax_x = dMax_x_in
    else:
        dMax_x = np.datetime64(np.nanmax(aTime_all), 'Y')
    if dMin_x_in is not None:
        dMin_x = dMin_x_in
    else:
        dMin_x = np.datetime64(np.nanmin(aTime_all), 'Y')

    if dMax_y_in is not None:
        dMax_y = dMax_y_in
    else:
        dMax_y = np.nanmax([aData_all, aData_upper_all, aData_lower_all])

    if dMin_y_in is not None:
        dMin_y = dMin_y_in
    else:
        # if it has negative value, change here
        dMin_y = np.nanmin([aData_all, aData_upper_all, aData_lower_all])

    if (dMax_y <= dMin_y):
        return
    else:
        iFlag_force_limit = 1
        if iFlag_force_limit == 1:
            pass
        else:
            dMin_y = dMin_y - 0.10 * (dMax_y-dMin_y)
            dMax_y = dMax_y + 0.25 * (dMax_y-dMin_y)  # leave space for legend

    if dSpace_y_in is not None:
        iFlag_space_y = 1
        dSpace_y = dSpace_y_in
    else:
        iFlag_space_y = 1
        dSpace_y = (dMax_y - dMin_y) / 4.0
        if dSpace_y < 1:
            pass
        else:
            dSpace_y = int(dSpace_y)
        pass

    if iFlag_miniplot_in is not None:
        iFlag_miniplot = iFlag_miniplot_in
        # set up location and range
        # dMin_mini_x = dMin_x + (dMax_x-dMin_x) * 0.7
        # dMax_mini_x =  dMax_x - (dMax_x-dMin_x) * 0.1
        # dMin_mini_y = dMin_y + (dMax_y-dMin_y) * 0.34
        # dMax_mini_y = dMax_y - (dMax_y-dMin_y) * 0.51

        dMin_mini_x = dMin_x + (dMax_x-dMin_x) * 0.7
        dMax_mini_x = dMax_x - (dMax_x-dMin_x) * 0.1
        dMin_mini_y = dMin_y + (dMax_y-dMin_y) * 0.05
        dMax_mini_y = dMax_y - (dMax_y-dMin_y) * 0.65
    else:
        iFlag_miniplot = 0

    fig = plt.figure(dpi=iDPI)
    fig.set_figwidth(iSize_x)
    fig.set_figheight(iSize_y)

    left, width = 0.1, 0.8
    bottom, height = 0.1, 0.5
    dY_mini = 0.60
    dX_mini = 0.15
    width_mini = 0.28
    heigh_mini = 0.40
    rect_full = [left, bottom, width, height]
    rect_mini = [dY_mini, dX_mini, width_mini, heigh_mini]

    ax_full = plt.axes(rect_full)
    if iFlag_miniplot == 1:
        ax_mini = plt.axes(rect_mini)
        ax_all = [ax_full, ax_mini]
    else:
        ax_all = [ax_full]

    nYear = int((dMax_x-dMin_x) / 5)
    pYear = mpl.dates.YearLocator(nYear)   # every year
    pYear_min = mpl.dates.YearLocator(1)   # every year
    pMonth = mpl.dates.MonthLocator()  # every month
    pMonth_min = mpl.dates.MonthLocator(6)  # every month

    if sDate_type_in is not None:
        if sDate_type_in == 'month':
            pMonth = mpl.dates.MonthLocator(3)
        else:
            print(sDate_type_in)
    else:
        pass

    if sFormat_y_in is not None:
        iFlag_format_y = 1
        sFormat_y = sFormat_y_in
    else:
        iFlag_format_y = 0
        sFormat_y = '{:.1f}'

    if sLocation_legend_in is not None:
        sLocation_legend = sLocation_legend_in
    else:
        sLocation_legend = "upper right"

    if aLocation_legend_in is not None:
        aLocation_legend = aLocation_legend_in
    else:
        aLocation_legend = (1.0, 1.0)

    if ncolumn_in is not None:
        ncolumn = ncolumn_in
    else:
        ncolumn = 1

    sYear_format = mpl.dates.DateFormatter('%Y')

    # start loop for each data

    for iax in range(len(ax_all)):
        ax = ax_all[iax]
        ax.tick_params(direction='in', top=True, right=True)
        if iax == 0:
            aLegend_artist = []
            aLabel = []
        else:
            pass
        for i in np.arange(1, nData+1):
            x1 = aTime_all[i-1]
            y1 = aData_all[i-1]
            y1_upper = aData_upper_all[i-1]
            y1_lower = aData_lower_all[i-1]
            sc_var = ax.fill_between(x1,
                                     y1_upper,
                                     y1_lower,
                                     alpha=0.5, color=aColor_fill[i-1])

            tsp, = ax.plot(x1, y1,
                           color=aColor[i-1], linestyle=aLinestyle[i-1],
                           label=aLabel_legend[i-1])

            if iax == 0:
                aLegend_artist.append(tsp)
                aLabel.append(aLabel_legend[i-1])

                # calculate linear regression
                iFlag_trend = aFlag_trend[i-1]
                if iFlag_trend == 1:
                    nan_index = np.where(y1 == missing_value)
                    y1[nan_index] = np.nan
                    good_index = np.where(~np.isnan(y1))
                    x_dummy = np.array([i.timestamp() for i in x1])
                    x_dummy = x_dummy[good_index]
                    y_dummy = y1[good_index]
                    coef = np.polyfit(x_dummy, y_dummy, 1)
                    poly1d_fn = np.poly1d(coef)
                    mn = np.min(x_dummy)
                    mx = np.max(x_dummy)
                    x2 = [mn, mx]
                    y2 = poly1d_fn(x2)
                    x2 = [datetime.fromtimestamp(i) for i in x2]
                    ax.plot(x2, y2, color='orange',
                            linestyle='-.',  linewidth=0.5)

        # unqiue setting
        if iax == 0:
            ax.axis('on')
            ax.grid(which='major', color='grey', linestyle='--', axis='y')
            ax.xaxis.set_major_locator(pYear)
            ax.xaxis.set_minor_locator(pMonth)
            ax.xaxis.set_major_formatter(sYear_format)
            ax.tick_params(axis="x", labelsize=10)
            ax.tick_params(axis="y", labelsize=10)
            ax.set_xmargin(0.05)
            ax.set_ymargin(0.15)
            ax.set_xlabel('Year', fontsize=12)
            ax.set_title(sTitle, loc='center', fontsize=15)
            ax.set_xlim(dMin_x, dMax_x)
            ax.set_ylabel(sLabel_y, fontsize=12)

            if iFlag_log == 1:
                aLabel_y = list()
                if dSpace_y >= 1:
                    dSpace_y = int(dSpace_y)
                    nlabel = int((dMax_y - dMin_y) / dSpace_y) + 1
                    for i in np.arange(0, nlabel, 1):
                        ii = int(dMin_y) + i * dSpace_y
                        sTicklabel = r'$10^{{{}}}$'.format(int(ii))
                        aLabel_y.append(sTicklabel)
                        pass
                    ticks = np.arange(0, nlabel, 1) * dSpace_y + int(dMin_y)
                    ax.set_yticks(ticks)
                    ax.set_yticklabels(aLabel_y)
                    pass
                else:
                    nlabel = int((dMax_y - dMin_y) / dSpace_y) + 1
                    for i in np.arange(0, nlabel, 1):
                        ii = int(dMin_y) + i * dSpace_y
                        iii = sFormat_y.format(ii)
                        sTicklabel = r'$10^{{{}}}$'.format(iii)
                        aLabel_y.append(sTicklabel)
                        pass
                    ticks = np.arange(0, nlabel, 1) * dSpace_y + dMin_y
                    ax.set_yticks(ticks)
                    ax.set_yticklabels(aLabel_y)
                    pass

                ax.yaxis.set_minor_locator(mpl.ticker.AutoMinorLocator())
                pass
            else:
                if iFlag_scientific_notation == 1:
                    formatter = mpl.ticker.ScalarFormatter(useMathText=True)
                    formatter.set_scientific(True)
                    ax.yaxis.set_major_formatter(formatter)
                    pass
                else:
                    if (iFlag_space_y == 0):
                        ax.yaxis.set_major_locator(
                            mpl.ticker.MaxNLocator(prune='upper',    nbins=5))
                    else:
                        ax.yaxis.set_major_locator(
                            mpl.ticker.MultipleLocator(dSpace_y))
                        ax.yaxis.set_minor_locator(mpl.ticker.AutoMinorLocator())
                        pass

                    if (iFlag_format_y == 1):
                        sFormat_y_dummy = sFormat_y.replace("{", "{x")
                        ax.yaxis.set_major_formatter(
                            mpl.ticker.StrMethodFormatter(sFormat_y_dummy))
                        pass
                    else:
                        pass

                    pass

            if (iReverse_y == 1):  # be careful here
                ax.set_ylim(dMax_y, dMin_y)
            else:
                ax.set_ylim(dMin_y, dMax_y)

            ax.legend(aLegend_artist, aLabel, bbox_to_anchor=aLocation_legend,
                      loc=sLocation_legend, fontsize=12, ncol=ncolumn)
        else:
            if iFlag_log == 1:
                aLabel_y = list()
                dSpace_y_mini = dSpace_y / 3
                if dSpace_y_mini >= 1:
                    dSpace_y_mini = int(dSpace_y_mini)
                    nlabel = int((dMax_y - dMin_y) / dSpace_y_mini) + 1
                    for i in np.arange(0, nlabel, 1):
                        ii = int(dMin_y) + i * dSpace_y_mini
                        sTicklabel = r'$10^{{{}}}$'.format(int(ii))
                        aLabel_y.append(sTicklabel)
                        pass
                    ticks = np.arange(0, nlabel, 1) * \
                        dSpace_y_mini + int(dMin_y)
                    ax.set_yticks(ticks)
                    ax.set_yticklabels(aLabel_y)
                else:
                    nlabel = int((dMax_y - dMin_y) / dSpace_y_mini) + 1
                    for i in np.arange(0, nlabel, 1):
                        ii = int(dMin_y) + i * dSpace_y_mini
                        iii = sFormat_y.format(ii)
                        sTicklabel = r'$10^{{{}}}$'.format(iii)
                        aLabel_y.append(sTicklabel)
                        pass
                    ticks = np.arange(0, nlabel, 1) * dSpace_y_mini + dMin_y
                    ax.set_yticks(ticks)
                    ax.set_yticklabels(aLabel_y)
                    pass
                ax.yaxis.set_minor_locator(mpl.ticker.AutoMinorLocator())
                pass
            else:
                if iFlag_scientific_notation == 1:
                    formatter = mpl.ticker.ScalarFormatter(useMathText=True)
                    formatter.set_scientific(True)
                    ax.yaxis.set_major_formatter(formatter)
                    pass
                else:
                    dSpace_y_mini = dSpace_y / 4
                    ax.yaxis.set_major_locator(
                        mpl.ticker.MultipleLocator(dSpace_y_mini))
                    ax.yaxis.set_minor_locator(mpl.ticker.AutoMinorLocator())
                    if (iFlag_format_y == 1):
                        sFormat_y_dummy = sFormat_y.replace("{", "{x")
                        ax.yaxis.set_major_formatter(
                            mpl.ticker.StrMethodFormatter(sFormat_y_dummy))
                    pass
            ax.xaxis.set_major_locator(pYear_min)
            ax.xaxis.set_minor_locator(pMonth_min)
            ax.xaxis.set_major_formatter(sYear_format)
            ax.set_xlim(dMin_mini_x, dMax_mini_x)
            if (iReverse_y == 1):
                ax.set_ylim(dMax_mini_y, dMin_mini_y)
            else:
                ax.set_ylim(dMin_mini_y, dMax_mini_y)
            pass

    plt.savefig(sFilename_out, bbox_inches='tight')
    plt.close('all')
    plt.clf()



import numpy as np
import datetime
import pyearth.toolbox.date.julian as julian
from pyearth.toolbox.date.day_in_month import day_in_month


def convert_time_series_daily_to_monthly(aData_daily_in,
                                         iYear_start_in,
                                         iMonth_start_in,
                                         iDay_start_in,
                                         iYear_end_in,
                                         iMonth_end_in,
                                         iDay_end_in,
                                         sType_in=None):
    """
    Convert a daily time series data into a monthly data

    Args:
        aData_daily_in (numpy.array): The daily time series data
        iYear_start_in (int): The start year
        iMonth_start_in (int): The start month
        iDay_start_in (int): The start day
        iYear_end_in (int): The end year
        iMonth_end_in (int): The end month
        iDay_end_in (int): The end day
        sType_in (string, optional): Method for aggregation. Defaults to None.

    Returns:
        numpy.array: aData_monthly_out
    """
    if sType_in is not None:
        sType = sType_in
    else:
        sType = 'mean'

    # check the length of data

    aData_daily = np.array(aData_daily_in)
    aData_monthly_out = list()

    dummy1 = datetime.datetime(iYear_start_in, iMonth_start_in, iDay_start_in)
    # dummy2 = datetime.datetime(iYear_end, iMonth_end, iDay_end)
    lJulian_start = julian.to_jd(dummy1, fmt='jd')
    # julian2 = julian.to_jd(dummy2, fmt='jd')
    # the old package
    # lJulian_start = gcal2jd(iYear_start_in, iMonth_start_in, iDay_start_in)
    for iYear in range(iYear_start_in, iYear_end_in + 1):
        if iYear == iYear_start_in:
            iMonth_start = iMonth_start_in
        else:
            iMonth_start = 1

        if iYear == iYear_end_in:
            iMonth_end = iMonth_end_in
        else:
            iMonth_end = 12

        for iMonth in range(iMonth_start, iMonth_end+1):
            # get all the data from this year-month

            if sType == "mean":

                if iYear == iYear_start_in and iMonth == iMonth_start_in:
                    iDay_start = iDay_start_in
                else:
                    iDay_start = 1

                if iYear == iYear_end_in and iMonth == iMonth_end_in:
                    iDay_end = iDay_end_in
                else:
                    iDay_end = day_in_month(iYear, iMonth)

                dummy = 0.0
                count = 0
                for iDay in range(iDay_start, iDay_end+1):
                    dummy2 = datetime.datetime(iYear, iMonth, iDay)
                    lJulian = julian.to_jd(dummy2, fmt='jd')
                    # lJulian=gcal2jd(iYear, iMonth, iDay)
                    dummy_index = int(lJulian-lJulian_start)
                    dummy = dummy + aData_daily[dummy_index]
                    count = count + 1
                    pass

                aData_monthly_out.append(dummy/count)

                pass
            else:
                if sType == 'sum':
                    if iYear == iYear_start_in and iMonth == iMonth_start_in:
                        iDay_start = iDay_start_in
                    else:
                        iDay_start = 1

                    if iYear == iYear_end_in and iMonth == iMonth_end_in:
                        iDay_end = iDay_end_in
                    else:
                        iDay_end = day_in_month(iYear, iMonth)

                    dummy = 0.0

                    for iDay in range(iDay_start, iDay_end+1):
                        # lJulian=gcal2jd(iYear, iMonth, iDay)
                        dummy2 = datetime.datetime(iYear, iMonth, iDay)
                        lJulian = julian.to_jd(dummy2, fmt='jd')
                        dummy_index = int(lJulian-lJulian_start)
                        dummy = dummy + aData_daily[dummy_index]
                        pass

                    aData_monthly_out.append(dummy)
                    pass
                else:
                    pass

            pass

    aData_monthly_out = np.array(aData_monthly_out)
    return aData_monthly_out



# convert_wgs_to_utm function, see https://stackoverflow.com/a/40140326/4556479
import math
def convert_wgs_to_utm(lon: float, lat: float):
    """Based on lat and lng, return best utm epsg-code"""
    utm_band = str((math.floor((lon + 180) / 6 ) % 60) + 1)
    if len(utm_band) == 1:
        utm_band = '0'+utm_band
    if lat >= 0:
        epsg_code = '326' + utm_band
        return epsg_code
    epsg_code = '327' + utm_band
    return epsg_code


import os
import numpy as np
from osgeo import gdal, ogr, osr

def rasterize_vector(sFilename_vector_in, sFilename_raster_out, 
                     dResolution_x, dResolution_y, 
                     iFlag_boundary_only_in = None,
                     iFlag_use_field_value_in=None,
                     sAttribute_name_in=None,
                     dMissing_value_in=None,
                     dField_value_in=None,
                     dFill_value_in=None,
                     iDataType_out=None,
                     dMin_x_in =None, dMax_x_in=None, dMin_y_in=None, dMax_y_in=None, 
                     nRow_in=None, nColumn_in=None,
                     pProjection_target_in = None ):
    """
    Converts any shapefile to a raster
    :param sFilename_vector_in: STR of a shapefile name (with directory e.g., "C:/temp/poly.shp")
    :param sFilename_raster_out: STR of target file name, including directory; must end on ".tif"
    :param pixel_size: INT of pixel size (default: 10)
    :param no_data_value: Numeric (INT/FLOAT) for no-data pixels (default: -9999)
    :param iDataType_out: gdal.GDALDataType raster data type - default=gdal.GDT_Float32 (32 bit floating point)
    :kwarg field_name: name of the shapefile's field with values to burn to the raster
    :return: produces the shapefile defined with sFilename_vector_in
    """

    # open data source
    try:
        pDatasource_vector = ogr.Open(sFilename_vector_in)
    except RuntimeError as e:
        print("Error: Could not open %s." % str(sFilename_vector_in))
        return None
    
    # Create a new vector layer for the boundary
    pDatasource_boundary = ogr.GetDriverByName('Memory').CreateDataSource('out')
    
    if os.path.exists(sFilename_raster_out):
        os.remove(sFilename_raster_out)
    
    if sAttribute_name_in is not None:        
        sAttribute_name = sAttribute_name_in

    if iFlag_use_field_value_in is None:
        iFlag_use_field_value = 0
    else:
        iFlag_use_field_value = 1
    
    if dField_value_in is None:
        dField_value = 1
    else:
        dField_value = dField_value_in

    if iFlag_boundary_only_in is None:
        iFlag_boundary_only = 1
    else:
        iFlag_boundary_only = iFlag_boundary_only_in
    
    if dFill_value_in is None:
        dFill_value = 1
    else:
        dFill_value = dFill_value_in
    
    pLayer_vector = pDatasource_vector.GetLayer()
    pSpatialRef_source = pLayer_vector.GetSpatialRef()
    pProjection_source = pSpatialRef_source.ExportToWkt()
    if pProjection_target_in is None:
        pProjection_target = pProjection_source
    else:
        pProjection_target = pProjection_target_in  
    
    pSpatialRef_target = osr.SpatialReference()
    pSpatialRef_target.ImportFromWkt(pProjection_target)
    
    if iDataType_out is None:
        iDataType = gdal.GDT_Float32 #gdal.GDT_Int16
    else:
        iDataType = iDataType_out

    if dMissing_value_in is None:
        dMissing_value = -9999
    else:
        dMissing_value = dMissing_value_in    

    if dMin_x_in is None or dMax_x_in is None or dMin_y_in is None or dMax_y_in is None:
        # read extent
        dMin_x, dMax_x, dMin_y, dMax_y = pLayer_vector.GetExtent()
    else:
        dMin_x = dMin_x_in
        dMax_x = dMax_x_in
        dMin_y = dMin_y_in
        dMax_y = dMax_y_in       

    #check the geometry type is polygon or not
    pFeature = pLayer_vector.GetNextFeature()
    pGeometry = pFeature.GetGeometryRef()
    if pGeometry.GetGeometryName() == 'POLYGON':
        pass
    else:
        if pGeometry.GetGeometryName() == 'MULTIPOLYGON':
            print('This is multipolygon')
        else:
            iFlag_boundary_only = 1    
     
    if nRow_in is not None:
        nrow = nRow_in
    else:
        nrow = int((dMax_y - dMin_y) / dResolution_y)    
    if nColumn_in is not None:
        ncolumn = nColumn_in
    else:
        ncolumn = int((dMax_x - dMin_x) / dResolution_x)

    if nrow == 0 or ncolumn == 0:
        #print('Error: resolution is too high')
        return

    #get raster drive
    pRaster_Driver = gdal.GetDriverByName('GTiff')
    pDatasource_raster = pRaster_Driver.Create(sFilename_raster_out, ncolumn, nrow, 1, eType=iDataType)
    pDatasource_raster.SetGeoTransform((dMin_x, dResolution_x, 0, dMax_y, 0, -dResolution_y))
    band = pDatasource_raster.GetRasterBand(1)    
    band.Fill(dMissing_value)
    band.SetNoDataValue(dMissing_value)

    # get spatial reference system and assign to raster
    pDatasource_raster.SetProjection(pSpatialRef_target.ExportToWkt())

    if iFlag_use_field_value == 1:
        if iFlag_boundary_only == 0:
            pLayer_boundary = pDatasource_boundary.CreateLayer('boundary', srs=pSpatialRef_source)
            geometry = pFeature.GetGeometryRef()        
            pFeature_boundary = ogr.Feature(pLayer_boundary.GetLayerDefn()) 
            pFeature_boundary.SetGeometry(geometry.Boundary())
            pLayer_boundary.CreateFeature(pFeature_boundary)
            gdal.RasterizeLayer(pDatasource_raster, [1], pLayer_vector, None, None, burn_values=[dField_value],
                    options=["ALL_TOUCHED=TRUE"])                
            gdal.RasterizeLayer(pDatasource_raster, [1], pLayer_boundary, None, None,
                options=["ALL_TOUCHED=TRUE"])
        else:
            gdal.RasterizeLayer(pDatasource_raster, [1], pLayer_vector, None, None, 
                    options=["ALL_TOUCHED=TRUE", "ATTRIBUTE=" + sAttribute_name])
    else: #use user provided value for field, 
        if iFlag_boundary_only == 0: #fill the polygon with a different value
            pLayer_boundary = pDatasource_boundary.CreateLayer('boundary', srs=pSpatialRef_source)
            geometry = pFeature.GetGeometryRef()        
            pFeature_boundary = ogr.Feature(pLayer_boundary.GetLayerDefn()) 
            pFeature_boundary.SetGeometry(geometry.Boundary())
            pLayer_boundary.CreateFeature(pFeature_boundary)
            gdal.RasterizeLayer(pDatasource_raster, [1], pLayer_vector, None, None, burn_values=[dFill_value],
                    options=["ALL_TOUCHED=TRUE"])                
            gdal.RasterizeLayer(pDatasource_raster, [1], pLayer_boundary, None, None, burn_values=[dField_value],
                options=["ALL_TOUCHED=TRUE"])
        else:
            gdal.RasterizeLayer(pDatasource_raster, [1], pLayer_vector, None, None, burn_values=[dField_value],
                    options=["ALL_TOUCHED=TRUE"])

    # release raster band
    band.FlushCache()
    pDatasource_boundary.Destroy()
    pDatasource_raster = None
    pSpatialRef_source = None
    pSpatialRef_target = None

    return


import numpy as np
from pyearth.gis.location.convert_longitude_latitude_to_sphere_3d import convert_longitude_latitude_to_sphere_3d

#https://stackoverflow.com/questions/8204998/how-to-check-if-a-pointlonc-latc-lie-on-a-great-circle-running-from-lona-lata

def calculate_distance_to_plane(dLongitude1_in, dLatitude1_in, 
                                dLongitude2_in, dLatitude2_in, 
                                dLongitude3_in, dLatitude3_in, 
                                iFlag_radian = None):
    
    if iFlag_radian is None:
        dLongitude1_radian_in, dLatitude1_radian_in = np.radians(np.array((dLongitude1_in, dLatitude1_in) ))
        dLongitude2_radian_in, dLatitude2_radian_in = np.radians(np.array((dLongitude2_in, dLatitude2_in) )) #this is the middle one
        dLongitude3_radian_in, dLatitude3_radian_in = np.radians(np.array((dLongitude3_in, dLatitude3_in) ))
        pass
    else:         
        dLongitude1_radian_in, dLatitude1_radian_in = dLongitude1_in , dLatitude1_in  
        dLongitude2_radian_in, dLatitude2_radian_in = dLongitude2_in , dLatitude2_in #this is the middle one
        dLongitude3_radian_in, dLatitude3_radian_in = dLongitude3_in , dLatitude3_in
        pass
        
    # The points in 3D space
    x1,y1,z1 = convert_longitude_latitude_to_sphere_3d(dLongitude1_radian_in, dLatitude1_radian_in)
    x2,y2,z2 = convert_longitude_latitude_to_sphere_3d(dLongitude2_radian_in, dLatitude2_radian_in)
    x3,y3,z3 = convert_longitude_latitude_to_sphere_3d(dLongitude3_radian_in, dLatitude3_radian_in)
    #The formula is x+b*y+c*z=0 
    
    c = (-x1*y3 + x3* y1)/( z1*y3 - z3*y1 )
    b = (  -x1*z3 + x3 * z1 ) / (y1 * z3 - y3*z1)
    distance = abs(  x2 + b * y2 + c * z2 )
    return distance


import os
import json
import numpy as np
from osgeo import osr, gdal, ogr
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import matplotlib.cm as cm
from matplotlib import animation
import cartopy.crs as ccrs
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
from pyearth.gis.location.get_geometry_coordinates import get_geometry_coordinates

pProjection = ccrs.PlateCarree()  # for latlon data only
iFigwidth_default = 8
iFigheight_default = 6

def animate_vector_polygon_data(
    sFilename_mesh_in,
    sFilename_animation_json_in,
    sFilename_animation_out,
    iFlag_type_in=None,
    iFigwidth_in=None,
    iFigheight_in=None,
    aExtent_in=None,
    sTitle_in = None,
    pProjection_map_in=None):
    
    if iFigwidth_in is None:
        iFigwidth_in = iFigwidth_default

    if iFigheight_in is None:
        iFigheight_in = iFigheight_default

    if iFlag_type_in is None:
        iFlag_type_in = 1

    dLat_min = 90
    dLat_max = -90
    dLon_min = 180
    dLon_max = -180    
       
    aCell_animation = list()
    aData = list()     
    aData_raw = list()
    aLongitude = list()
    aLatitude = list()
    aAVertex = list()
    #get domain range
    pDriver = ogr.GetDriverByName("GeoJSON")
    pDataset = pDriver.Open(sFilename_mesh_in, gdal.GA_ReadOnly)
    pLayer = pDataset.GetLayer(0)
    pSrs = osr.SpatialReference()
    pSrs.ImportFromEPSG(4326)  # WGS84 lat/lon
    for pFeature in pLayer:
        pGeometry_in = pFeature.GetGeometryRef()
        sGeometry_type = pGeometry_in.GetGeometryName()
        if sGeometry_type == "POLYGON":            
            aCoords_gcs = get_geometry_coordinates(pGeometry_in)
            dLon_max = np.max([dLon_max, np.max(aCoords_gcs[:, 0])])
            dLon_min = np.min([dLon_min, np.min(aCoords_gcs[:, 0])])
            dLat_max = np.max([dLat_max, np.max(aCoords_gcs[:, 1])])
            dLat_min = np.min([dLat_min, np.min(aCoords_gcs[:, 1])])

    if pProjection_map_in is not None:
        pProjection_map = pProjection_map_in
    else:
        pProjection_map = ccrs.Orthographic(
            central_longitude=0.50 * (dLon_max + dLon_min),
            central_latitude=0.50 * (dLat_max + dLat_min),
            globe=None,
        )

    fig = plt.figure(dpi=150)
    fig.set_figwidth(iFigwidth_in)
    fig.set_figheight(iFigheight_in)
    ax = fig.add_axes([0.1, 0.1, 0.65, 0.8], projection=pProjection_map)
    ax.set_global()    
    marginx = (dLon_max - dLon_min) / 20
    marginy = (dLat_max - dLat_min) / 20
    if aExtent_in is None:
        aExtent = [
            dLon_min - marginx,
            dLon_max + marginx,
            dLat_min - marginy,
            dLat_max + marginy,
        ]
    else:
        aExtent = aExtent_in

    ax.set_extent(aExtent)
    ax.coastlines()  # resolution='110m')
    if iFlag_type_in == 1:  # full
        gl = ax.gridlines(
            crs=ccrs.PlateCarree(),
            draw_labels=True,
            linewidth=1,
            color="gray",
            alpha=0.3,
            linestyle="--",
        )
        gl.xlabel_style = {"size": 8, "color": "k", "rotation": 0, "ha": "right"}
        gl.ylabel_style = {"size": 8, "color": "k", "rotation": 90, "weight": "normal"}
    else:  # track mode
        pass

    cmap = cm.get_cmap("Spectral")
    if sTitle_in is not None:
        # setting a title for the plot
        #sText = "Priority flood in HexWatershed"
        ax.text(
            0.5,
            1.08,
            sTitle_in,
            verticalalignment="center",
            horizontalalignment="center",
            transform=ax.transAxes,
            color="black",
            fontsize=9,
        )
    cmap_reversed = cmap.reversed()   

    #get dataset
    with open(sFilename_animation_json_in) as json_file:
        aCell_animation = json.load(json_file)
        ncell_animation = len(aCell_animation)

    for i in range(ncell_animation):
        pCell_animation = aCell_animation[i]        
        aData.append( float(pCell_animation["dElevation"]))
        aData_raw.append(float(pCell_animation["dElevation_raw"]))
        aLongitude.append(float(pCell_animation["dLongitude_center_degree"]))
        aLatitude.append(float(pCell_animation["dLatitude_center_degree"]))
        aAVertex.append(pCell_animation["vVertex"])

    #convert to numpy array
    aData = np.array(aData)
    aData_raw = np.array(aData_raw)
    aLongitude = np.array(aLongitude)
    aLatitude = np.array(aLatitude)
    dData_max = np.max(aData)
    dData_min = np.min(aData)

    sText = ""
    x1 = 0.0
    y1 = 0.0
    pArtist1 = ax.text(
        x1,
        y1,
        sText,
        verticalalignment="center",
        horizontalalignment="right",
        transform=ax.transAxes,
        color="black",
        fontsize=12,
    )
    x2 = 0.0
    y2 = 0.0
    pArtist2 = ax.text(
        x2,
        y2,
        sText,
        verticalalignment="center",
        horizontalalignment="left",
        transform=ax.transAxes,
        color="black",
        fontsize=12,
    )

    # trasform data
    norm = plt.Normalize(dData_min, dData_max)
    sm = plt.cm.ScalarMappable(cmap = cmap_reversed, norm=norm)
    sm.set_array(aData)
    fig.canvas.draw()
    # Section 2
    ax_pos = ax.get_position() # get the original position
    #use this ax to set the colorbar ax position
    ax_cb = fig.add_axes([ax_pos.x1+0.06, ax_pos.y0, 0.02, ax_pos.height])
    cb = fig.colorbar(sm, cax=ax_cb)
    sUnit = r"Unit: m"
    cb.ax.get_yaxis().set_ticks_position("right")
    cb.ax.get_yaxis().labelpad = 5
    cb.ax.set_ylabel(sUnit, rotation=90)
    cb.ax.get_yaxis().set_label_position("left")
    cb.ax.tick_params(labelsize=6)

    #calculate ahead
    aColor_index = (aData - dData_min) / (dData_max - dData_min)
    
    def animate(i):    
        # get the time step global id and updated elevation     
        dlon = aLongitude[i]
        dlat = aLatitude[i]
        dummy0 = aData_raw[i]
        dummy = aData[i]  
        aVertex = aAVertex[i]
        aLocation = np.array([[vertex["dLongitude_degree"], vertex["dLatitude_degree"]] for vertex in aVertex])
       
        rgb = cmap_reversed( aColor_index[i] )
        polygon = mpatches.Polygon(
            aLocation,
            closed=True,
            facecolor=rgb,
            edgecolor="none",
            transform=ccrs.Geodetic(),
        )
        pArtist0 = ax.add_patch(polygon)
        if iFlag_type_in == 1: #??
            dLon_min_zoom = dLon_min
            dLon_max_zoom = dLon_max
            dLat_min_zoom = dLat_min
            dLat_max_zoom = dLat_max
        else:
            dLon_min_zoom = dlon - marginx * 2
            dLon_max_zoom = dlon + marginx * 2
            dLat_min_zoom = dlat - marginy * 2
            dLat_max_zoom = dlat + marginy * 2

        x = (dlon - dLon_min_zoom) / (dLon_max_zoom - dLon_min_zoom)
        y = (dlat - dLat_min_zoom) / (dLat_max_zoom - dLat_min_zoom)
    
        if dummy0 > dummy:
            y1 = y + 0.05
            y2 = y - 0.05
        else:
            y1 = y - 0.05
            y2 = y + 0.05

        sText1 = "Before: " + "{:0.2f}".format(dummy0) + "m"
        pArtist1.set_x(x)
        pArtist1.set_y(y1)
        pArtist1.set_text(sText1)
        sText2 = "After: " + "{:0.2f}".format(dummy) + "m"
        pArtist2.set_x(x)
        pArtist2.set_y(y2)
        pArtist2.set_text(sText2)
        if iFlag_type_in == 2:
            aExtent_zoom = [
                dlon - marginx * 2,
                dlon + marginx * 2,
                dlat - marginy * 2,
                dlat + marginy * 2,
            ]
            ax.set_extent(aExtent_zoom)

        return pArtist0, pArtist1, pArtist2

    #remove if already exist  
    #check output file extension
    if os.path.exists(sFilename_animation_out):
        os.remove(sFilename_animation_out)
    if sFilename_animation_out.endswith('.gif'):
        plt.rcParams[
            "animation.convert_path"
        ] = "/share/apps/ImageMagick/7.1.0-52/bin/convert"
        writer = 'imagemagick'        
        anim = animation.FuncAnimation(fig, animate, frames=ncell_animation, interval=400, blit=False)
        anim.save(sFilename_animation_out, writer= writer)

    else: #use video instead      
        plt.rcParams[
            "animation.ffmpeg_path"
        ] = '/people/liao313/.conda/envs/hexwatershed/bin/ffmpeg' #'/share/apps/ffmpeg/6.1.1/bin/ffmpeg            
        #Writer = animation.writers['ffmpeg']
        #writer = Writer(fps=5, metadata=dict(artist='Chang Liao'), bitrate=-1)
        writer=animation.FFMpegWriter(fps=5, metadata=dict(artist='Chang Liao'), bitrate=-1)
        anim = animation.FuncAnimation(fig, animate, frames=ncell_animation, interval=1, blit=True)
        anim.save(sFilename_animation_out, writer=writer)
    return



import numpy as np
from datetime import datetime

import matplotlib as mpl
import matplotlib.pyplot as plt
from pyearth.system.define_global_variables import *


def plot_time_series_data_multiple_temporal_resolution_fill(aTime_all, 
                                                            aData_all,
                                                            sFilename_out,
                                                            iDPI_in=None,
                                                            iFlag_trend_in=None,
                                                            iReverse_Y_in=None,
                                                            iSize_X_in=None,
                                                            iSize_Y_in=None,
                                                            dMax_Y_in=None,
                                                            dMin_Y_in=None,
                                                            dSpace_y_in=None,
                                                            aMarker_in=None,
                                                            aColor_in=None,
                                                            aLinestyle_in=None,
                                                            sLabel_Y_in=None,
                                                            aLabel_legend_in=None,
                                                            sTitle_in=None):

    nData = len(aData_all)
    aData = [aData_all]
    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_trend_in is not None:
        iFlag_trend = 1
    else:
        iFlag_trend = 0

    if iReverse_Y_in is not None:
        iReverse_Y = 1
    else:
        iReverse_Y = 0

    if iSize_X_in is not None:
        iSize_X = iSize_X_in
    else:
        iSize_X = 12
    if iSize_Y_in is not None:
        iSize_Y = iSize_Y_in
    else:
        iSize_Y = 9

    if sLabel_Y_in is not None:
        sLabel_Y = sLabel_Y_in
    else:
        sLabel_Y = ''
    if aLabel_legend_in is not None:
        aLabel_legend = aLabel_legend_in
    else:
        aLabel_legend = {}
    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    if aMarker_in is not None:
        aMarker = aMarker_in
    else:
        aMarker = np.full(nData, '+')
    if aColor_in is not None:
        aColor = aColor_in
    else:
        aColor = np.full(nData, 'black')
    if aLinestyle_in is not None:
        aLinestyle = aLinestyle_in
    else:

        aLinestyle = np.full(nData, '-')

    if dMax_Y_in is not None:
        dMax_Y = dMax_Y_in
    else:
        dMax_Y = np.nanmax(aData) * 1.2
    if dMin_Y_in is not None:
        dMin_Y = dMin_Y_in
    else:
        dMin_Y = np.nanmin(aData)  # if it has negative value, change here
    if (dMax_Y <= dMin_Y):
        return

    if dSpace_y_in is not None:
        dSpace_y = dSpace_y_in
    else:
        pass

    fig = plt.figure(dpi=iDPI)
    fig.set_figwidth(iSize_X)
    fig.set_figheight(iSize_Y)
    ax = fig.add_axes([0.1, 0.5, 0.8, 0.4])
    pYear = mpl.dates.YearLocator(1)   # every year
    pMonth = mpl.dates.MonthLocator()  # every month
    sYear_format = mpl.dates.DateFormatter('%Y')

    for i in np.arange(1, nData+1):
        x1 = aTime_all[i-1]

        if i == 1:
            # we need to plot low and high fill
            y_top = aData_all[i-1][2]
            y_bot = aData_all[i-1][1]
            ax.fill_between(x1, y_top, y_bot,  facecolor='cornflowerblue')
            # plot mean
            y1 = aData_all[i-1][0]

            ax.plot(x1, y1,
                    color=aColor[i-1], linestyle=aLinestyle[i-1],
                    marker=aMarker[i-1],
                    label=aLabel_legend[i-1])

        else:
            y1 = aData_all[i-1]

            ax.plot(x1, y1,
                    color=aColor[i-1], linestyle=aLinestyle[i-1],
                    marker=aMarker[i-1],
                    label=aLabel_legend[i-1])

        # calculate linear regression
        nan_index = np.where(y1 == missing_value)
        y1[nan_index] = np.nan
        good_index = np.where(~np.isnan(y1))
        if iFlag_trend == 1:
            x_dummy = np.array([i.timestamp() for i in x1])
            x_dummy = x_dummy[good_index]
            y_dummy = y1[good_index]
            coef = np.polyfit(x_dummy, y_dummy, 1)
            poly1d_fn = np.poly1d(coef)
            mn = np.min(x_dummy)
            mx = np.max(x_dummy)
            x2 = [mn, mx]
            y2 = poly1d_fn(x2)
            x2 = [datetime.fromtimestamp(i) for i in x2]
            ax.plot(x2, y2, color='orange', linestyle='-.',  linewidth=0.5)

    ax.axis('on')
    ax.grid(which='major', color='grey', linestyle='--', axis='y')

    # ax.set_aspect(dRatio)  #this one set the y / x ratio
    ax.xaxis.set_major_locator(pYear)
    ax.xaxis.set_minor_locator(pMonth)
    ax.xaxis.set_major_formatter(sYear_format)
    ax.tick_params(axis="x", labelsize=10)
    ax.tick_params(axis="y", labelsize=10)

    ax.set_xmargin(0.05)
    ax.set_ymargin(0.15)

    ax.set_xlabel('Year', fontsize=12)
    ax.set_ylabel(sLabel_Y, fontsize=12)
    ax.set_title(sTitle, loc='center', fontsize=15)
    # round to nearest years...
    x_min = np.datetime64(aTime_all[1][0], 'Y')
    x_max = np.datetime64(aTime_all[1][-1], 'Y') + np.timedelta64(1, 'Y')
    ax.set_xlim(x_min, x_max)
    if dMax_Y < 1000 and dMax_Y > 0.001:
        ax.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
    else:
        ax.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1e'))

    ax.yaxis.set_major_locator(mpl.ticker.MultipleLocator(dSpace_y))

    if (iReverse_Y == 1):
        ax.set_ylim(dMax_Y, dMin_Y)
    else:
        ax.set_ylim(dMin_Y, dMax_Y)
    ax.legend(bbox_to_anchor=(1.0, 1.0), loc="upper right", fontsize=12)
    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')
    plt.clf()
    # print('finished plotting')





def retrieve_usgs_site_information_nwis(sSiteId):
    try:
        import requests
    except ImportError as e:
        raise ImportError(
            "The package 'requests' is required for this function to run.") from e
    
    squaremile2squarekm = 2.58999
    
    nwis_site_url = f"https://waterservices.usgs.gov/nwis/site/?format=rdb&siteOutput=Expanded&sites={sSiteId}"
    response = requests.get(nwis_site_url)
    pData = response.text       
    dLon = None
    dLat = None
    dDrainage = None
    
    for line in pData.splitlines():
        if line.startswith("#"):
            #print(line)
            continue
        data = line.split("\t")
        if len(data) >= 4:
            #print(data)
            if data[0] == 'agency_cd':
                index_lon = data.index('dec_long_va')
                index_lat = data.index('dec_lat_va')
                index_drai = data.index('drain_area_va')      
            
            if data[0] == 'USGS' and data[1] == sSiteId:
                dLon = float(data[index_lon])
                dLat = float(data[index_lat])
                dDrainage= float(data[index_drai]) * squaremile2squarekm
    
    return dLon, dLat, dDrainage


def line_count(sFilename_in):
    """
    Count the line count in a text file

    Args:
        sFilename_in (string): The filename

    Returns:
        int: The total line count
    """
    ifs=open(sFilename_in, 'rb') 
    i=0 
    sLine0=(ifs.readline())#.rstrip()
    sLine= sLine0.decode("utf-8", 'ignore')
    while len(sLine) > 0:
        i = i+1
        sLine0=(ifs.readline())
        sLine= sLine0.decode("utf-8", 'ignore')
        
        
    return i


import numpy as np
def get_geometry_coordinates(geometry):
    
    sGeometry_type = geometry.GetGeometryName()
    if sGeometry_type =='POINT':
        return get_point_coords(geometry)
    elif sGeometry_type == 'LINESTRING':
        return get_linestring_coords(geometry)
    elif sGeometry_type =='POLYGON':
        return get_polygon_exterior_coords(geometry)
    else:
        raise ValueError("Unsupported geometry type.")

def get_polygon_exterior_coords(polygon_geometry):
    exterior_coords = []
    ring = polygon_geometry.GetGeometryRef(0)  # Get the exterior ring
    for i in range(ring.GetPointCount()):
        point = ring.GetPoint(i)
        exterior_coords.append((point[0], point[1]))
    return np.array(exterior_coords)

def get_linestring_coords(linestring_geometry):
    coords = []
    for i in range(linestring_geometry.GetPointCount()):
        point = linestring_geometry.GetPoint(i)
        coords.append((point[0], point[1]))
    return np.array(coords)

def get_point_coords(point_geometry):
    point = point_geometry.GetPoint()
    return np.array([(point[0], point[1])])



import datetime
import pyearth.toolbox.date.julian as julian
def day_of_year(iYear_in, iMonth_in, iDay_in):
    """
    Calculate the day in a year

    Args:
        iYear_in (int): The year
        iMonth_in (int): The month
        iDay_in (int): The day

    Returns:
        int: The day in a year
    """

    if iYear_in is None:
        print("Parameter 'iYear_in' is undefined ")
        return -1
    if iMonth_in is None:
        print("Parameter 'iMonth_in' is undefined ")
        return -1
    else:
       if iMonth_in < 1 or iMonth_in > 12:
          print("Parameter 'iMonth_in' is not within correct range ")
          return -1
    if iDay_in is None:
        print("Parameter 'iDay_in' is undefined ")
        return -1
    else:
       if iDay_in < 1 or iDay_in > 31:
          print('Parameter iDay_in is not within correct range')
          return -1

    dummy1 = datetime.datetime(iYear_in, 1, 1)
    lJulian_start = julian.to_jd(dummy1, fmt='jd')

    dummy1 = datetime.datetime(iYear_in, iMonth_in, iDay_in)
    lJulian_end = julian.to_jd(dummy1, fmt='jd')

    dayofyear = int (lJulian_end[1] - lJulian_start[1] ) + 1

    return dayofyear




def slurm_update_checkpoint_file(iIndex_restart,
                                 iIndex_start,
                                 iIndex_end,
                                 sFilename_checkpoint):
    """
    Write and save the checkpoint information

    Args:
        iIndex_restart (int): [description]
        iIndex_start (int): [description]
        iIndex_end (int): [description]
        sFilename_checkpoint (string): [description]
    """
    ofs = open(sFilename_checkpoint, "w")  # write mode
    sLine = "{:0d}".format(iIndex_restart) + '\n'
    ofs.write(sLine)
    sLine = "{:0d}".format(iIndex_start) + '\n'
    ofs.write(sLine)
    sLine = "{:0d}".format(iIndex_end) + '\n'
    ofs.write(sLine)
    ofs.close()

    return




#from math import acos
import numpy as np
def calculate_angle_between_vectors_degrees(u, v):
    """Return the angle between two vectors in any dimension space,
    in degrees.
    """
    a = np.dot(u, v)
    b = np.linalg.norm(u)
    c = np.linalg.norm(v)
    d = a / (b * c)
    if d > 1:
        d = 1
    if d < -1:
        d = -1

    #e = acos(d) 
    e = np.arccos(d)
    
    f = np.degrees(e)

    return f


import numpy as np
def convert_longitude_latitude_to_sphere_3d(dLongitude_in, dLatitude_in, iFlag_radian= None):
    """Convert a point given latitude and longitude in radians to
    3-dimensional space, assuming a sphere radius of one."""

    if iFlag_radian is None:
        dLongitude_radian, dLatitude_radian = np.radians([dLongitude_in, dLatitude_in])
    else:
        dLongitude_radian, dLatitude_radian = dLongitude_in, dLatitude_in
        pass

    a = np.cos(dLatitude_radian) * np.cos(dLongitude_radian)
    b = np.cos(dLatitude_radian) * np.sin(dLongitude_radian)
    c = np.sin(dLatitude_radian)
    d = np.array((a,b,c))
    
    return d


def leap_year(iYear_in):
    """
    Check whether this year is a leap year or not.

    Args:
        iYear_in (int): The year

    Returns:
        bool:  True or False
    """
    if iYear_in % 400 == 0:
        return True
    if iYear_in % 100 == 0:
        return False
    if iYear_in % 4 == 0:
        return True
    else:
        return False



from osgeo import gdal, ogr, osr, gdalconst
def gdal_get_raster_extent(sFilename_in):
    ds = gdal.Open(sFilename_in, gdal.GA_ReadOnly)
    gt = ds.GetGeoTransform()
    min_x = gt[0]
    max_x = gt[0] + gt[1]*ds.RasterXSize
    min_y = gt[3] + gt[5]*ds.RasterYSize
    max_y = gt[3]
    ds = None
    return min_x, max_x, min_y, max_y


def create_diverge_rgb_color_hex(ncolor, iFlag_reverse_in=None):
    """
    choose diverge color from: https://colorbrewer2.org/
    """

    if ncolor < 3 or ncolor > 12:
        return -1
    else:
        if ncolor == 3:
            colors_hex = ['#fc8d59',
                          '#ffffbf',
                          '#91bfdb']

        else:
            if ncolor == 4:
                colors_hex = ['#d7191c',
                              '#fdae61',
                              '#abdda4',
                              '#2b83ba']
            if ncolor == 5:
                colors_hex = ['#d7191c',
                              '#fdae61',
                              '#ffffbf',
                              '#abdda4',
                              '#2b83ba']
            if ncolor == 6:
                colors_hex = ['#66c2a5',
                              '#fc8d62',
                              '#8da0cb',
                              '#e78ac3',
                              '#a6d854',
                              '#ffd92f']

                # another option
                colors_hex = ['#e41a1c',
                              '#377eb8',
                              '#4daf4a',
                              '#984ea3',
                              '#ff7f00',
                              '#ffff33']

            if ncolor == 7:
                colors_hex = ['#d73027',
                              '#fc8d59',
                              '#fee090',
                              '#ffffbf',
                              '#e0f3f8',
                              '#91bfdb',
                              '#4575b4']
            if ncolor == 8:
                colors_hex = ['#d53e4f',
                              '#f46d43',
                              '#fdae61',
                              '#fee08b',
                              '#e6f598',
                              '#abdda4',
                              '#66c2a5',
                              '#3288bd']
            if ncolor == 9:
                colors_hex = ['#d53e4f',
                              '#f46d43',
                              '#fdae61',
                              '#fee08b',
                              '#ffffbf',
                              '#e6f598',
                              '#abdda4',
                              '#66c2a5',
                              '#3288bd']
            if ncolor == 10:
                colors_hex = ['#9e0142',
                              '#d53e4f',
                              '#f46d43',
                              '#fdae61',
                              '#fee08b',
                              '#e6f598',
                              '#abdda4',
                              '#66c2a5',
                              '#3288bd',
                              '#5e4fa2']
            else:
                pass

        # add the reverse feature
        if iFlag_reverse_in is not None:
            colors_hex.reverse()
        else:
            pass
        return colors_hex



import math as Math
import numpy as np


def calculate_ticks_space(aData, nstep_in=5, iFlag_small_in=None):
    """
    Computes domain with given nstep_in encompassing series aData
    @ params
    aData    - Required - A list-like object of integers or floats
    nstep_in - Optional - Tick frequency
    """

    if iFlag_small_in is not None:
        iFlag_small = 1
    else:
        iFlag_small = 0
    if iFlag_small == 1:
        xMax = np.nanmax(aData)
        xMin = np.nanmin(aData)
        dSpace = (xMax - xMin) / (nstep_in-2)
        dMax = xMax * 1.2
        dMin = xMin * 0.8

    else:

        xMax = Math.ceil(np.nanmax(aData))
        xMin = Math.floor(np.nanmin(aData))

        dSpace = (xMax - xMin) / (nstep_in-2)

        dMax = xMax + Math.ceil(dSpace)
        dMin = Math.floor(xMin - dSpace)
    return (dSpace, dMin, dMax)
